#!/usr/bin/env python
from __future__ import division
#from sage.all import *    #WARNING!!!! If using only the Explicit Failure Rate computation method, do NOT enable SAGE. Slows things down. Use simple Python instead.
from math import factorial as fac
from math import erf, floor, ceil, sqrt, pi, log, exp, sinh 
from fractions import gcd
import os,subprocess,time,datetime
import operator as op
from sympy import isprime,nextprime
from collections import defaultdict
from ast import literal_eval
log_infinity=999999999

#----------------  utilities.py (Reference:https://github.com/pq-crystals/kyber/tree/master/scripts)  -----------------


def gaussian_center_weight(sigma, t):
    """ Weight of the gaussian of std deviation s, on the interval [-t, t]
    :param x: (float)
    :param y: (float)
    :returns: erf( t / (sigma*\sqrt 2) )
    """
    return erf(t / (sigma * sqrt(2.)))


def binomial(x, y):
    """ Binomial coefficient
    :param x: (integer)
    :param y: (integer)
    :returns: y choose x
    """    
    try:
        binom = fac(x) // fac(y) // fac(x - y)    #note the use of // instead of /. // returns the int floor value of the division op, making sure that no OverflowErrors occur.
    except ValueError:
        binom = 0
    return binom


def centered_binomial_pdf(k, x):
    """ Probability density function of the centered binomial law of param k at x
    :param k: (integer)
    :param x: (integer)
    :returns: p_k(x)
    """    
    return binomial(2*k, x+k) / 2.**(2*k)


def build_centered_binomial_law(k):
    """ Construct the binomial law as a dictionnary
    :param k: (integer)
    :param x: (integer)
    :returns: A dictionnary {x:p_k(x) for x in {-k..k}}
    """    
    D = {}
    for i in range(-k, k+1):
        D[i] = centered_binomial_pdf(k, i)
    return D


def mod_switch(x, q, rq):
    """ Modulus switching (rounding to a different discretization of the Torus)
    :param x: value to round (integer)
    :param q: input modulus (integer)
    :param rq: output modulus (integer)
    """    
    return int(round(1.* rq * x / q) % rq)


def mod_centered(x, q):
    """ reduction mod q, centered (ie represented in -q/2 .. q/2)
    :param x: value to round (integer)
    :param q: input modulus (integer)
    """        
    a = x % q
    if a < q/2:
        return a
    return a - q


def build_mod_switching_error_law(q, rq):
    """ Construct Error law: law of the difference introduced by switching from and back a uniform value mod q
    :param q: original modulus (integer)
    :param rq: intermediate modulus (integer)
    """        
    D = {}
    V = {}
    for x in range(q):
        y = mod_switch(x, q, rq)
        z = mod_switch(y, rq, q)
        d = mod_centered(x - z, q)
        D[d] = D.get(d, 0) + 1./q
        V[y] = V.get(y, 0) + 1

    return D

#Create probability distribution for a sparse, ternary distribution with Hamming weight h. +1, -1 each occur with probability h/2n, 0 occurs with probability 1-(h/n).
def build_sparse_ternary_law(n,hamm_wt):
    D = {}
    domain=[-1,0,1]
    D[domain[0]] = float(hamm_wt/(2*n))
    D[domain[1]] = float(1.0 - (hamm_wt/n))
    D[domain[2]] = float(hamm_wt/(2*n))
    return D

#Scale both the domain and corresponding probabilites in a error law by the given factor
def scale_law(law,scale_factor):
    C = {}
    for item in law:
        c = scale_factor*item
        C[c] = C.get(c, 0) + (scale_factor*law[item])
    return C


def law_convolution(A, B):
    """ Construct the convolution of two laws (sum of independent variables from two input laws)
    :param A: first input law (dictionnary)
    :param B: second input law (dictionnary)
    """        

    C = {}
    for a in A:
        for b in B:
            c = a+b
            C[c] = C.get(c, 0) + A[a] * B[b]
    return C


def law_product(A, B):
    """ Construct the law of the product of independent variables from two input laws
    :param A: first input law (dictionnary)
    :param B: second input law (dictionnary)
    """        
    C = {}
    for a in A:
        for b in B:
            c = a*b
            C[c] = C.get(c, 0) + A[a] * B[b]
    return C


def clean_dist(A):
    """ Clean a distribution to accelerate further computation (drop element of the support with proba less than 2^-300)
    :param A: input law (dictionnary)
    """        
    B = {}
    for (x, y) in A.items():
        if y>2**(-300):
            B[x] = y
    return B


def iter_law_convolution(A, i):
    """ compute the -ith forld convolution of a distribution (using double-and-add)
    :param A: first input law (dictionnary)
    :param i: (integer)
    """    
    D = {0: 1.0}
    i_bin = bin(i)[2:]  # binary representation of n
    for ch in i_bin:
        D = law_convolution(D, D)
        D = clean_dist(D)
        if ch == '1':
            D = law_convolution(D, A)
            D = clean_dist(D)
    return D


def tail_probability(D, t):
    '''
    Probability that an drawn from D is strictly greater than t in absolute value
    :param D: Law (Dictionnary)
    :param t: tail parameter (integer)
    '''
    s = 0
    ma = max(D.keys())
    if t >= ma:
        return 0
    for i in reversed(range(int(ceil(t)), ma)):  # Summing in reverse for better numerical precision (assuming tails are decreasing)
        s += D.get(i, 0) + D.get(-i, 0)
    return s


# Binary search on a non-decreasing function. Adapted from John Schank's PARI GP code.
# func(x) non-decreasing; maxLT() finds maximum x such that f(x) <= target
# Parameter func is a lambda function, defined in and passed as parameter from the calling function.
def maxLT(func,target,low,high):
    #print "\nmaxLT(), func is",func,"target is",target,"low is",low,"high is",high
    if(low >= (high-1)):
        return low
    midpt = ceil((low+high)/2)
    if(func(midpt)<=target):
        #print "maxLT(), func(midpt) is",func(midpt),"<=target. Calling maxLT(func,target,midpt,high)"
        return maxLT(func,target,midpt,high)
    else:
        #print "maxLT(), func(midpt) is",func(midpt),">target. Calling maxLT(func,target,low,midpt)"
        return maxLT(func,target,low,midpt)
    # endif


#-------------------------------------------------  secanalysis_functions.py  ------------------------------------------------------------


#Calculate log base 2 of an input
def log2(n):
    res=float(log(n)/log(2))
    return res


#Simulate the binomial() function of SAGE, i.e., nCr.
def my_binomial(n,r):
    r = min(r, n-r)
    if r == 0: return 1
    numer = reduce(op.mul, xrange(n, n-r, -1))
    denom = reduce(op.mul, xrange(1, r+1))
    return numer//denom


#Check whether the input q differs at most a value th from a power of two
def close_to_power_of_two(q,th):
    exp  = int(floor(log2(q)))
    power = 2**exp
    ok = 0
    if ((q < power+th) and (q > power - th)):
        ok = 1
    return ok


#Calculate the root Hermite factor for a given BKZ blocksize
def get_rhf(b):
    if((b>=2) and (b<50)):    #BKZ with block-size=2 is exactly the LLL algorithm
        return 1.0219    #best known practical value of LLL's root Hermite factor. Theoretical upper bound of LLL's RHF is 1.0754. See Gama's "Predicting Lattice Reduction".
    else:    #BKZ is the generalization of LLL to block-sizes>2. Relation to obtain RHF from b only makes sense for b>50
        return ( (pi*b)**(1./b) * b / (2*pi*exp(1)))**(1./(2.*b-2.))


## log_2 of best plausible Quantum Cost of SVP in dimension b. For random lattices, hence has an extra 'b' multiplicative factor as compared to Ideal lattices.
def svp_quantum_nist(b,ring,cost):
    if(ring==0):
        return (float(log2(b)) + float(b*cost))	#Cosr for random lattices, note the additional log b
    else:
        return float(b*cost)


#Encapsulate ALL bandwidth-related data of a parameter set
#Note: PST represents the (R)LWR-based Philips Security Technologies framework for PKE and KEM. Whether it is CPA or CCA secure depends on the failure probability. FP<2^(-128) signifies it is CCA secure.
class Bandwidth(object):
    def __init__(self,bw_frodo,bw_kybercpakem,bw_kyberccakem,bw_round2,round2_pk_sz,round2_ct_sz):
        self._bw_frodo=float(bw_frodo)
        self._bw_pstkem=float(bw_round2)
        self._round2_pk_sz=float(round2_pk_sz)
        self._round2_ct_sz=float(round2_ct_sz)
        self._bw_kybercpakem=float(bw_kybercpakem)
        self._bw_kyberccakem=bw_kyberccakem    #all bandwidth data in BYTES!
        self._gain_pstkem_frodo=float(bw_round2/bw_frodo)
        self._gain_pstkem_kybercpakem=float(bw_round2/bw_kybercpakem)
        self._gain_pstkem_kyberccakem=float(bw_round2/bw_kyberccakem)
    def __repr__(self):
        return "Bandwidth (Frodo): %s B, Bandwidth (Kyber CPA KEM): %s B, Bandwidth (Kyber CCA KEM): %s B, \nBandwidth (Round2): %s B, Round2.public-key size: %s B, Round2.ciphertext size: %s B, \nRatio (PST.KEM/Frodo): %s, Ratio (PST/Kyber.CPA): %s, Ratio (PST/Kyber.CCA): %s"%(self._bw_frodo,self._bw_kybercpakem,self._bw_kyberccakem,self._bw_pstkem,self._round2_pk_sz,self._round2_ct_sz,self._gain_pstkem_frodo,self._gain_pstkem_kybercpakem,self._gain_pstkem_kyberccakem)


#Calculate the value of (R)LWR rounding modulus so that rounding error \approx LWE Gaussian error of S.D. sigma
def get_lwrp(q,sigma):
    log_p=log2(q)-(1+(0.5*log2(3.0))+log2(sigma))
    log_p=floor(log_p)            #p should be a power of 2, and should be the smaller choice to keep rounding error as large as possible.
    return float(2**log_p)


#Calculate the bandwidth in B for Round2.KEM and Round2.PKE
#Parameter indsectype determines which one out of Round2.KEM or Round2.PKE we are currently considering.
def bandwidth_pstkem(dim,q,p,t,n_bar,m_bar,key_size,B,indsectype):
    tau_sz = 8.0    #1 byte for representing \tau, the type of f_n permutation function used to generate A
    sigma_sz = key_size    #Length of the seed \sigma used as input to f_n, has same length as the shared-secret length
    B_sz = float(dim*n_bar*ceil(log2(p)))    #size of B, the public-key component of the initiator/decryptor.
    round2_pk_sz = tau_sz + sigma_sz + B_sz

    U_sz = dim*m_bar*ceil(log2(p))    #size of U, the ciphertext component of the responder/encryptor.
    v_sz = float((key_size*ceil(log2(t)))/B)    #size of v, the ciphertext component of the responder/encryptor.
    round2_cpa_ct_sz = U_sz + v_sz    #size of a Round2.KEM ciphertext in bits.
    if(indsectype==0):    #IND-CPA security targeted. So we are considering Round2.KEM.
        bw_cpakem = ceil(round2_pk_sz/8.) + ceil(round2_cpa_ct_sz/8.)    #Convert to bytes
        return (int(ceil(bw_cpakem)), int(ceil(round2_pk_sz/8.)), int(ceil(round2_cpa_ct_sz/8.)))    #round upwards to the closest byte
    elif(indsectype==1):    #IND-CCA security targeted. We are considering Round2.PKE.
        round2_cca_ct_sz = round2_cpa_ct_sz + key_size    #represents the component g in the CCA-PKE ciphertext, has same length as the shared secret.
        round2_cca_ct_sz += (16*8)    #represents the AESGCM authentication tag, size=16 bytes
        round2_cca_ct_sz += (12*8)    #represents the AESGCM initiatilization vector, size=12 bytes
        bw_ccapke = ceil(round2_pk_sz/8.) + ceil(round2_cca_ct_sz/8.)    #convert to bytes
        return (int(ceil(bw_ccapke)), int(ceil(round2_pk_sz/8.)), int(ceil(round2_cca_ct_sz/8.)))    #round upwards to the closest byte


#Return bandwidth (bytes) of the FRODO LWE-based KEX, values taken from their parameter sets, converted to bytes (*1024).
def get_bw_frodo(seclevel):
    if(seclevel==0):
        return 14.22*1024.            #Taken from Frodo's parameters
    elif(seclevel==1):
        return 22.57*1024.
    elif(seclevel==2):
        return 25.93*1024.
    else:
        return -1.0               #Undefined security level


#Create a new bandwidth object, encapsulating all bandwidth performance data and comparison with Frodo for the current parameter set.
def get_bw(seclevel,q,p,n,t,n_bar,m_bar,key_size,B,indsectype):
    (bw_round2, round2_pk_sz, round2_ct_sz) = bandwidth_pstkem(n,q,p,t,n_bar,m_bar,key_size,B,indsectype)
    bw_kybercpakem=2080.0    #Taken from the Kyber paper. Sum of PK and ciphertext in bytes for the CPA version of Kyber.
    bw_kyberccakem=2272.0    #Taken from the Kyber paper. Sum of PK and ciphertext. In bytes.
    bw_obj=Bandwidth(get_bw_frodo(seclevel),bw_kybercpakem,bw_kyberccakem,bw_round2,round2_pk_sz,round2_ct_sz)
    return bw_obj


#Calculate the value of B (key bits extracted per matrix component) for a given n_bar,m_bar, satisfying B.n_bar.m_bar=keylength.
def get_B(alice_instances,bob_instances,keylength):
    return int(ceil(keylength/(alice_instances*bob_instances)))


#Class representing a set of valid, secure parameters
class Param(object):
    def __init__(self,sec_level,theta,ring,q,p,error_sd,n_maxunsafe,n_minsafe,keylength,keystrength,key_size,t,B,n_bar,m_bar,expl_fail_pessimistic,expl_fail_compromise,expl_fail_quasioptimistic,expl_fail_optimistic,dual_best_cost,dual_best_b,dual_best_m,dual_attack_gap,hybrid_mitmcost,hybrid_cost_lr,hybrid_r,use_sufficient_decryptionsuccess,indsectype):
        self._sec_level=sec_level
        self._q=float(q)
        self._p=float(p)
        self._sigma=float(error_sd)
        self._theta=float(theta)
        self._n_maxunsafe=n_maxunsafe
        if(ring==0):
            self._dim=n_minsafe
            self._n=1
        else:
            self._dim=n_minsafe
            self._n=n_minsafe
        bw_obj=get_bw(sec_level,q,p,self._dim,t,n_bar,m_bar,keylength,B,indsectype)
        self._bw=bw_obj
        self._keylength_nist=keylength
        self._keystrength_nist=keystrength
        self._t=t                #Compression space of ciphertext in PST.KEM
        self._B=B                #Space in which the key of PST.KEM, i.e., the plaintext of PST.PKE lives.
        self._n_bar=n_bar        #Alice LWE/R instances
        self._m_bar=m_bar        #Bob LWE/R instances
        self._expl_fail_pessimistic=expl_fail_pessimistic
        self._expl_fail_compromise=expl_fail_compromise
        self._expl_fail_quasioptimistic=expl_fail_quasioptimistic
        self._expl_fail_optimistic=expl_fail_optimistic
        self._keylength=key_size
	self._primal_best_cost = 0
	self._primal_best_b = 0
	self._primal_best_m = 0
	self._dual_best_cost = dual_best_cost
	self._dual_best_b = dual_best_b
	self._dual_best_m = dual_best_m
        self._dual_attack_gap = dual_attack_gap    #Security gap representing the difference in solving Decision LWR via the dual attack, versus the cost of solving Search LWE by reducing it to Decision LWR using the Bogdanov reduction, Theorem 3.
	self._hybrid_mitmcost = hybrid_mitmcost
	self._hybrid_best_cost = hybrid_cost_lr
	self._hybrid_r = hybrid_r
        self._albrechtattack_cost = 0
        self._albrechtattack_k_opt = 0
        self._primal_best_cost_pt = 0
        self._dual_best_cost_pt = 0
        self._hybrid_best_cost_pt = 0
        self._albrechtattack_cost_pt = 0
        self._use_sufficient_decryptionsuccess = use_sufficient_decryptionsuccess
    def __repr__(self):
        if(self._sec_level==0):
            level="Classical"
        elif(self._sec_level==1):
            level="Post-quantum"
        elif(self._sec_level==2):
            level="Paranoid"
        else:
            level="Undefined"
        return "Security level: %s, Keylength_nist: %s Keystrength_nist: %s\nPrimal_cost: %s, Primal_b: %s, Primal_m: %s\nDual_cost: %s, Dual_b: %s, Dual_m: %s\nHybrid_cost: %s, Hybrid_cost_MITM: %s, Hybrid_r: %s\nAlbrecht's sparse-secret attack cost: %s, Optimal number of columns in A to drop: %s,\n q: %s, p: %s, \nlog2(q): %s, log2(p): %s, \nLWR noise standard deviation: %s,\nn_maxunsafe: %s, dim: %s, n: %s,\n%s,\nKeylength: %s,\nt: %s, B: %s, n_bar: %s, m_bar: %s, theta: %s \nKEX Failure probability is: exactly 2^(%s) (Explicit,pessimistic), exactly 2^(%s) (Explicit,compromise), exactly 2^(%s) (Explicit, quasi-optimistic), exactly 2^(%s) (Explicit,optimistic),\nUsing PST sufficient decryption-success condition:%s."%(level,self._keylength_nist,self._keystrength_nist, self._primal_best_cost, self._primal_best_b, self._primal_best_m, self._dual_best_cost, self._dual_best_b, self._dual_best_m, self._hybrid_best_cost, self._hybrid_mitmcost, self._hybrid_r, self._albrechtattack_cost, self._albrechtattack_k_opt, self._q, self._p,log2(self._q), log2(self._p), self._sigma, self._n_maxunsafe, self._dim, self._n,self._bw, self._keylength, self._t, self._B,self._n_bar,self._m_bar, self._theta, self._expl_fail_pessimistic, self._expl_fail_compromise,self._expl_fail_quasioptimistic,self._expl_fail_optimistic,self._use_sufficient_decryptionsuccess)


#Caluclate the norm of the i-th Gram-Schmidt vector in a reduced q-ary lattice basis of dimension d, using the Geometric Series Assumption.
#Lattice may be scaled by a scaling factor to exploit small secrets
def gram_schmidt_norm(d,m,q,rhf,i,omega):
    return float((q**(m/d)) * (omega**((d-m-1)/d)) * (rhf**(d-(2*i)-1)))


#Calculate the log of the scaling/weighting factor used to build weighted lattices for attacks on small-secret LWE
def log_scaling_factor(sigma,theta,is_hnf=False,is_binary=False,is_ternary=False,is_sparsehnf=False,is_spbinary=False,is_spternary=True):
    if(is_hnf):
        log_w = 0.0
    if(is_binary):
        log_w = float(1.0+log2(sigma))             #Ignoring the 2.pi Gaussian normalization factor that Lizard uses, advised by Ronald.
        #0.5+log2(sigma)-(0.5*log2(pi))
    if(is_ternary):
        log_w = float(log2(sigma)+(0.5*(log2(3.0) - 1)))
        #log2(sigma)-1+(0.5*(log2(3)-log2(pi)))    #Ignoring the 2.pi factor in Lizard
    if(is_sparsehnf):
        log_w = 0.0             #TODO: Confirm if this is correct!!
    if(is_spbinary):
        log_w = float(log2(sigma)-(0.5*(log2(theta) + log2(1.0-theta))))
        #log2(sigma)-(0.5*(log2(2*pi) + log2(theta) + log2(1-theta)))    #Ignoring the 2.pi factor in Lizard
    if(is_spternary):
        log_w = float(log2(sigma)-(0.5*log2(theta)))
        #log2(sigma)-(0.5*(1+log2(pi)+log2(theta)))    #Ignoring the 2.pi factor in Lizard
    return log_w


#Calculate the log of the scaling factor specifically for the Dual attack, computed by balancing norms of secret and error per-vector, and not per-component.
def log_dual_scaling_factor(sigma,n,m,hamm_wt,is_hnf=False,is_binary=False,is_ternary=False,is_sparsehnf=False,is_spbinary=False,is_spternary=True):
    if(is_spternary):
        log_w = float( log2(sigma)+(0.5*(log2(m)-log2(hamm_wt))) )
    else:
        print "log_dual_scaling_factor(): Undefined secret type. Abort."
        log_w = 0
    return log_w


#Calculate the norm of the unique shortest vector (s,e,1) in the Primal Embedding lattice
def minNorm(n,m,sigma,theta,omega,is_hnf=False,is_binary=False,is_ternary=False,is_sparsehnf=False,is_spbinary=False,is_spternary=True):
    h=floor(n*theta)
    norm=0.0
    if(is_hnf):
        norm=sigma*sqrt(n+m)
    if(is_binary):
        norm=sqrt(float((omega**2)*(n/2)) + float(m*(sigma**2)))
    if(is_ternary):
        norm=sqrt(float(((omega**2)*2*n)/3) + float(m*(sigma**2)))
    if(is_sparsehnf):
        norm=sigma*sqrt(h+m)
    if(is_spbinary):
        norm=sqrt(float((omega**2)*h) + float(m*(sigma**2)))
    if(is_spternary):
        norm=sqrt(float((omega**2)*h) + float(m*(sigma**2)))
    return float(norm)


#Calculate the norm of the projection \widetilde(v) of the vector v=(s,e,1) on the vector space spanned by the last b-Gram-Schmidt vectors of the embedding lattice.
#v_norm: The norm of v=(s,e,1); v_dim: its dimension.
def projectedNorm(v_norm,b,v_dim):
    return float(v_norm*sqrt(float(b/v_dim)))


#Calculate the log of the volume of the lattice into which (primal) embedding is done, scaled or otherwise by a scaling/weighting factor omega.
#Calculating in log and not absolute values because q^d is huge and might quickly go out of hand.
def log_embedLatticeVol(q,d,n,omega):
    log_vol=float((d-n-1)*(log2(q))) + float(n*log2(omega))
    return log_vol
    

def print_secret_type(file,theta,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary):
    if(is_hnf):
        s="Current secret type is Gaussian (Hermite-Normal form)."
    if(is_binary):
        s="Current secret type is uniform binary."
    if(is_ternary):
        s="Current secret type is uniform ternary."
    if(is_sparsehnf):
        s="Current secret type is sparse Gaussian (Hermite-Normal form) with sparseness: "+str(theta)
    if(is_spbinary):
        s="Current secret type is sparse binary with sparseness:"+str(theta)
    if(is_spternary):
        s="Current secret type is sparse ternary with sparseness: "+str(theta)
    s=s+"\r\n\n"
    file.write(s)
    return 1


#Success condition of Primal attack against (R)LWE. Returns true if attack succeeds for given parameters and false otherwise.    
def primal_attack_success(sd,b,q,n,m,theta,is_hnf=False,is_binary=False,is_ternary=False,is_sparsehnf=False,is_spbinary=False,is_spternary=True):
    #Dimension of the lattice considered in the Primal attack.
    d=m+n+1    
    #Root-Hermite factor for a BKZ lattice reduction algorithm running with given block-size b.
    log_rhf=log2(get_rhf(b))    
    #Log2 of scaling factor for rescaling the lattice in case of small-secrets.
    log_omega=log_scaling_factor(sd,theta,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary)
    omega=float(2**log_omega)    
    #Norm of the unique shortest vector (s,e,1) that the primal attack searches for.
    norm_v=minNorm(n,m,sd,theta,omega,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary)
    dim_v=m+n+1
    #Norm of the projection of (s,e,1) on the span of the last b Gram-Schmidt vectors
    proj_norm_v=projectedNorm(norm_v,b,dim_v)    
    #Log2 of volume of the lattice considered in the Primal attack, possibly scaled due to small-secrets.
    log_latt_vol=log_embedLatticeVol(q,d,n,omega)    
    #Check embedding success condition
    lhs=log2(proj_norm_v)
    rhs=float((b+b-d)*log_rhf) + float(log_latt_vol/d)
    if(lhs<=rhs):
        #Primal attack succeeded
        return True
    else:
        #Primal attack failed
        return False


#for a given q,p,n calculate the cost of running the primal attack (if it succeeds), otherwise return infinity
def primalcost(b,q,p,n,m,theta,sec_level,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary,ring,bkzcost):
    #Calcuate the root Hermite factor
    rhf=get_rhf(b)
    #Calculate the standard deviation of the LWR rounding error
    sigma=float(q/(2*sqrt(3.0)*p))
    #Calculate primal attack success condition    
    if(primal_attack_success(sigma,b,q,n,m,theta,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary)):
        #Primal attack succeeds. Calculate the actual cost of running it, assuming one SVP call
	return svp_quantum_nist(b,ring,bkzcost)
    else:
        return log_infinity     #Primal attack fails. Cost of running it is undefinably large.


#Calculation of the distinguishing advantage of the Dual Attack and the corresponding attack cost, in log_2
#Note: the parameter "use_lwr_reduction" determines whether the (practical) dual attack cost (of solving Decision LWR) is calculated, 
#or the (theoretical) dual attack cost (of solving Search LWE by reducing it to Decision LWR, 
#via the Bogdanov reduction, Theorem 3 - https://eprint.iacr.org/2015/769.pdf) is calculated.
def dualcost(b,q,p,n,m,theta,sec_level,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary,ring,bkzcost,use_lwr_reduction):
    #Dimension of the dual lattice
    d=m+n    
    #Root-Hermite factor for a BKZ lattice reduction algorithm running with given block-size b.
    rhf=get_rhf(b)
    #Calculate the standard deviation of the LWR rounding error
    sd=float(q/(2*sqrt(3.0)*p))
    #Log2 of scaling factor for rescaling the lattice in case of small-secrets.
    log_omega=log_dual_scaling_factor(sd,n,m,int(floor(theta*n)),is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary)
    omega=float(2**log_omega)
    #Norm of the shortest vector in the scaled dual lattice. omega is the scaling factor calculated above.
    norm_v=(rhf**(d-1)) * ((q/omega)**(1. * n/d))
    tau= norm_v * sd / q
    log2_eps = -0.5 + (- 2 * pi * pi * tau * tau) / log(2)
    svp_cost=svp_quantum_nist(b,ring,bkzcost)
    total_cost = max(0, - 2 * log2_eps - svp_quantum_nist(b,ring,0.2075)) + svp_cost
    
    security_gap=0.0    #Default value
    #Sanity check. Currently, no efficient reductions to DECISION Ring-LWR are known.
    if((use_lwr_reduction==1) and (ring!=0)):
        print "Error.Cannot calculate Ring-LWE advantage based on Ring-LWR advantage. No efficient reductions for Decision-RLWR known. Abort!"
        exit()
    elif((use_lwr_reduction==1) and (ring==0)):
        #print "Calculating cost of solving Search-LWE by reducing to Decision-LWR."
        E=abs(float((2.**(log2_eps))/(4*q*m)) - float(2.**(n-(m*log2(p)))))    #Since E is getting squared, can disregard the negative sign to prevent log() from throwing UNDEFINED errors.
        if(E==0):
            log2_eps_lwe=0
        else:
            log_E_sq=2*log2(E)
            log2_eps_lwe=log_E_sq - m
        total_cost_lwe = max(0, - 2 * log2_eps_lwe - svp_quantum_nist(b,ring,0.2075)) + svp_cost
        security_gap=total_cost_lwe-total_cost
    return (total_cost,security_gap)


## Hybrid meet-in-the-middle and lattice reduction attack analysis ##


# Calculate the cost of guessing (either via Grover's search or via meet-in-the-middle) the non-zero components of a vector of length r, whose non-zero components are distributed according to the probability distribution coeffDist.
def mitm_cost(coeffDist,r):
    # Can use either Grover's search (quantum approach) or MITM (classical approach) to search r coefficients
    # using 2^(0.5 * entropy(coeffDist)) queries
    plogp = 0
    for symbol in coeffDist:
        probab=coeffDist.get(symbol,0)
        plogp+=(probab * log2(probab))
    cost_mitm=floor(0.5*r*(-plogp))
    return cost_mitm


# Set the root Hermite factor requirement such that the last Gram Schmidt vector in the reduced lattice basis (of dimension d)
# has length greater than twice the maximum coefficient size in coeffDist, the distribution of the target vector's components. 
def hybrid_hermite(d,m,q,coeffDist,omega):
    #For a sparse, ternary distribution, maximum coefficient size is 1.
    max_coeff_sz = 1
    min_gs_vec=2*omega*max_coeff_sz
    rhf_min=float((float((m/d)*log2(q)) + float(((d-m-1)/d)*log2(omega)) - log2(min_gs_vec))/d)
    return float(2**rhf_min)


# Back-calculate to estimate the largest BKZ block-size that results in a given root Hermite factor
def hybrid_blocksz(dim,rhf):
    bs = maxLT(lambda x: (-1*get_rhf(x)), 
        -rhf, 2, dim)
    return (bs+1)


def hybrid_cost_estimate_nist(q,p,n,max_m,theta,coeffDist,r,sec_level, ring,cost):
    sigma=float(q/(2.*sqrt(3.)*p))
    omega=float(2**(log_scaling_factor(sigma,theta)))
    ## Find an optimal (m,RHF) pair satisfying the following two conditions: 
    ## 1. The first G.S. vector has norm at most q, i.e., this is a necessary condition that lattice reduction succeeded in reducing the lattice basis.
    ## 2. The last G.S. vector has norm greater than twice the largest coefficient in coeffDist, the distribution of the target vector's components.

    ## max (m, h). m s.t. b*_1 < q; h s.t. b*_{m+(n-k)} > s
    ## Notice that with different m's tried out, omega remains the same since it is a function of constant n,q,p,theta
    m_opt = maxLT(lambda m: gram_schmidt_norm(m+(n-r),m,q,hybrid_hermite(m+(n-r),m,q,coeffDist,omega),1,omega), 
        q, 2, max_m)
    d_opt=m_opt+(n-r)
    rhf_opt=hybrid_hermite(d_opt,m_opt,q,coeffDist,omega)
    blocksz_opt=hybrid_blocksz(d_opt,rhf_opt)
    cost_lr=svp_quantum_nist(blocksz_opt, ring,cost) 
    cost_MITM=mitm_cost(coeffDist,r)
    return cost_lr
    

# Calculate the optimal meet-in-the-middle dimension r that results in (almost) eqaluuy balanced meet-in-the-middle and lattice reduction costs.
# Parameter coeffDist is a dictionary representing the distribution of the secret distribution, containing a (symbol)->(probability) mapping.
def hybrid_tradeoff_nist(q,p,n,max_m,theta,coeffDist,sec_level, ring,cost):
    wt=int(floor(theta*n))
    r_opt = maxLT(lambda r: (mitm_cost(coeffDist,r) - hybrid_cost_estimate_nist(q,p,n,max_m,theta,coeffDist,r,sec_level, ring,cost)), 
        0, 0, n)
    return r_opt


def hybrid_attack_cost_r_nist(q,p,n,max_m,theta,sec_level,ring,bkzcost):
    #probability distribution for the coefficients of a sparse-ternary distribution, implemented as a Python dictionary
    coeffDist=build_sparse_ternary_law(n,int(floor(theta*n)))
    #calculate optimal meet-in-the-middle dimension by trying to balance MITM cost and lattice reduction cost in the hybrid attack.
    r=hybrid_tradeoff_nist(q,p,n,max_m,theta,coeffDist,sec_level,ring,bkzcost)
    cost_mitm=mitm_cost(coeffDist,r)
    cost_lr=hybrid_cost_estimate_nist(q,p,n,max_m,theta,coeffDist,r,sec_level,ring,bkzcost)
    return (cost_mitm,cost_lr,r)


## END Hybrid meet-in-the-middle and lattice reduction attack analysis ##


#for a given lattice instance, optimize over BKZ blocksize b and number of (R)LWR samples m to get the minimal primal or dual attack cost.
def optimizeattack_b_m(q,p,n,max_m,theta,sec_level,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary,attack_type,ring,bkzcost,use_lwr_reduction):
    if(sec_level>3):
        print "Undefined security level. Abort."
        return -1
    best_cost=log_infinity    #Initialize optimal attack cost to infinity.
    best_b=2    #Least possible value                  #Optimal BKZ blocksize that results in minimum attack cost.
    best_m=1    #Least possible value                  #Optimal number of LWR samples that results in minimum attack cost.
    b_min=60                  #Smallest value of BKZ blocksize that we consider.
    #Below value for the maximum value of b that can be considered as available to an attacker has been taken from NewHope. 
    #For spKEX parameter selection (only), it makes sense to fix b_max such that the corresponding BKZ workload is 128-bits. 
    #However, while optimizing over b,m it actually makes sense to check even higher values of b to make sure that no (strange) combinations of b,m result in a smaller cost.
    dual_attack_gap=0.0    #Security gap in the cost of solving Decision LWR via dual attack, versus the cost of solving Search LWE by reducing it to Decision LWR.
    b_max=(2*n)+max_m+1
    if(b_max<b_min):
        return (0,0,0)
    for b in range(b_min,b_max):
        current_cost = svp_quantum_nist(b,ring,bkzcost)
        if(current_cost>best_cost):
            break        #Cost of running the attack is higher than the (previously) best cost. No point checking any more values of b,m since minimal cost has already been found
        for m in range(max(1,(b-n)),max_m):
            if(attack_type==1):
                #Primal attack
                attack_cost=primalcost(b,q,p,n,m,theta,sec_level,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary,ring,bkzcost)
            elif(attack_type==2):
                #Dual attack
                (attack_cost,dual_attack_gap)=dualcost(b,q,p,n,m,theta,sec_level,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary,ring,bkzcost,use_lwr_reduction)
            else:
                #Undefined attack type
                attack_cost=log_infinity
            if(attack_cost<best_cost):
                #Have found a b,m pair for which the attack cost is even cheaper than what it was so far.
                best_cost=attack_cost
                best_b=b
                best_m=m
    if(attack_type==1):
        correct_cost=primalcost(best_b,q,p,n,best_m,theta,sec_level,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary,ring,bkzcost)
    else:
        (correct_cost,dual_attack_gap)=dualcost(best_b,q,p,n,best_m,theta,sec_level,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary,ring,bkzcost,use_lwr_reduction)
    if((best_b != 2) and (best_cost != correct_cost)):
        return (-1,-1,-1,-1)
    else:
        return (best_b,best_m,best_cost,dual_attack_gap)


#Albrecht sparse attack analysis. 


#total running time complexity in bits, for Albrecht's "Drop A vectors corresponding to zero-components in the secret vector" attack (naive embodiment)
#time_bkz: Cost of running a lattice reduction with BKZ-b on the given lattice of dimension (n-k).
def albrecht_attack_cost_naive(n,k,h,time_bkz):
    repetitions = log2((binomial(n,k)) // (binomial((n-h),k)))    #Must be an integer. We use // (equivalent to int(floor())) to get a conservative (defender viewpoint) cost estimate of the attack.
    repetitions = max(0,repetitions)    #Ignore negative values
    #print "albrecht_attack_cost_naive(). repetitions",repetitions,"time_bkz",time_bkz
    return repetitions + int(floor(time_bkz))


#total running time complexity in bits, for Albrecht's "Drop sparse columns" attack (adaptive embodiment: tries to recover from the error of mistakenly dropping a non-sparse column)
def albrecht_attack_cost_adaptv(n,k,h,time_bkz):
    j = 0
    repetitions = 0
    l = k - 1            #Maximum number of erroneoulsy dropped non-sparse columns
    checks = 0
    while(j < l):
        drop_probab_j = (binomial((n-h),(k-j)) * binomial(h,j)) // (binomial(n,k))
        repetitions += drop_probab_j
        checks += binomial(k,j) * (2**j)
        j += 1
    if(repetitions>=log_infinity or checks>=log_infinity):
        return log_infinity + int(floor(time_bkz))
    else:
	#print "repetitions",repetitions,"checks",checks
	if(repetitions!=0):
	    repetitions = checks // repetitions
	#else:    repetitions = 0        
	repetitions = max(0,repetitions)    #ignore negative values
	return repetitions + int(floor(time_bkz))


#Optimize Albrecht's sparse column dropping attack for a given n and hammwt.
def optimize_albrecht_spattack(q,p,n,max_samples,theta,current_sec_level,ring,bkzcost,use_lwr_reduction,is_hnf=False,is_binary=False,is_ternary=False,is_sparsehnf=False,is_spbinary=False,is_spternary=True):
    min_alb_cost=log_infinity
    hammwt=int(floor(theta*n))
    k_min = 0
    k_max = n-hammwt
    k_opt = 0
    attack_types=[1,2]    #we check for both the Primal and dual flavors of Albrecht's attack.
    for attack_type in attack_types:
        for k in (k_min, k_max):    #The number of vectors in A that the attacker can choose to ignore.
            #First try using the Primal attack flavor. Usually stronger than the Dual attack flavor.
            (b,m,cost_lr,dual_gap) = optimizeattack_b_m(q,p,n,max_samples,theta,current_sec_level,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary,attack_type,ring,bkzcost,use_lwr_reduction)
            if(min(albrecht_attack_cost_naive(n,k,hammwt,cost_lr), albrecht_attack_cost_adaptv(n,k,hammwt,cost_lr)) < min_alb_cost):    #Considering both the Naive and Adaptive embodiment of Albrecht's attack.
                min_alb_cost = min(albrecht_attack_cost_naive(n,k,hammwt,cost_lr), albrecht_attack_cost_adaptv(n,k,hammwt,cost_lr))
                k_opt=k
    return (min_alb_cost, k_opt) 


#Load values of k (ignored components) and total attack cost from given files
def visualize_albrecht_attack_k_cost(k_filename,albcost_filename):
    k_file=open(k_filename,'r')
    albcost_file=open(albcost_filename,'r')
    k_list=[]
    albcost_list=[]
    for line in k_file:
        line=line.replace(","," ")
        line=line.split()
        if line:
            k_list=[int(i) for i in line]
    for line in albcost_file:
        line=line.replace(","," ")
        line=line.split()
        if line:
            albcost_list=[float(i) for i in line]
    k_file.close()
    albcost_file.close()
    return (k_list,albcost_list)


## End analysis of Albrecht's A-vector dropping attack against sparse-LWE, sparse-LWR ##
## Operational correctness analysis, of failure probability. ##


#Calculate the "Delta" upper bound on the failure probability in spKEX, dictated by the use of Ludo's reconciliation method.
def spkex_delta_bound(q,p,B,t,h):
    b_h=floor(log2(t))
    if(((q%p)==0) and ((p%(2**(B+b_h+1)))==0)):
        return float(p/(2**(B+1))) - float(p/(2**(B+b_h+1)))
    else:
        #Reconciliation assumptions not met. Abort
        return log_infinity


#Calculate the maximum tolerable upper bound of the El Gamal decryption failure in PST
def pst_cpapke_delta_bound(q,p,B,t,h):
    #print "pst_cpapke_delta_bound(q,p,B,t,h): ",q,p,B,t,h
    d=gcd(q,p)
    pstcpapke_bound = float((p*q)/(d*(2**(B+1))))
    #print "pstcpapke_bound",pstcpapke_bound
    return pstcpapke_bound


# Calculate whether a given parameter set satisfies the "sufficient" condition for decryption success, i.e., the condition ensuring that decryption is ALWAYS successful.
def pst_sufficient_success(q,p,B,t,h,ring):
    d = gcd(q,p)
    if((q%p)==0):
        norm_unbiasing_err_qp = 0
    else:
        norm_unbiasing_err_qp = float(p/(2*d))    #represents the randomized error added to ensure no bias in rounding if p does not divide q.
    if(ring==0):
        hammwt_mul = 1
    else:
        hammwt_mul = 2
    maxnorm_E1_R = float(hammwt_mul * h * (norm_unbiasing_err_qp + float((q+p)/(2*d))))
    maxnorm_S_E2 = float(hammwt_mul * h * (norm_unbiasing_err_qp + float(q/(2*d))))
    maxnorm_E3 = float(q/(2*d)) + float((p*q)/(2*d*t)) + norm_unbiasing_err_qp
    maxnorm_E = maxnorm_E1_R + maxnorm_S_E2 + maxnorm_E3
    if(maxnorm_E < pst_cpapke_delta_bound(q,p,B,t,h)):
        return True    #decryption will always succeed.
    else:
        return False    #decryption might 'sometimes' fail.


def pst_cpapke_delta_bound_pessimistic(q,p,B,t,h):
    #print "pst_cpapke_delta_bound_pessimistic(q,p,B,t,h): ",q,p,B,t,h
    d=gcd(q,p)
    pessimistic_pstcpapke_bound = pst_cpapke_delta_bound(q,p,B,t,h) - float((q+p)/(2*d)) - float((p*q)/(2*d*t))
    if((q%p)!=0):    #if q is not a multiple of p, then unbiased rounding error must be added into the computation of the delta_bound. See decryptionerror.text for more details, Eq. 10.
        pessimistic_pstcpapke_bound -= float(p/(2*d))
    #print "pessimistic_pstcpapke_bound",pessimistic_pstcpapke_bound
    return pessimistic_pstcpapke_bound


def pst_cpapke_delta_bound_quasioptimistic(q,p,B,t,h):
    #print "\n pst_cpapke_delta_bound_quasioptimistic(q,p,B,t,h): ",q,p,B,t,h
    d=gcd(q,p)
    #print "quasioptimistic_pstcpapke_bound",(pst_cpapke_delta_bound(q,p,B,t,h) - float((p*q)/(2*d*t)))
    return pst_cpapke_delta_bound(q,p,B,t,h) - float((p*q)/(2*d*t))


#Generate the pessimistic failure probability of PST.CPA-PKE for a given decryption error distribution
def gen_failure_probability_pessimistic(error_distribution,ring,bound,key_size,B):
    if(ring!=0):
        fail_pessimistic = tail_probability(error_distribution, floor(bound/4.)) * 4.0    #Ring case, Failure=4 * T.P.(h-fold convolution of chi_E'_1, Delta_P/4).
    else:
        fail_pessimistic = tail_probability(error_distribution, floor(bound/2.)) * 2.0    #Nonring case. Failure=2 * T.P.(h-fold convolution of chi_E'_1, Delta_P/2).
    fail_pessimistic *= (key_size/B)    #Union bound over all key components/coefficients
    #print "log_fail_pessimistic",log(fail_pessimistic + 2.**(-300))/log(2)
    return log(fail_pessimistic + 2.**(-300))/log(2)
    

#Generate the compromise failure probability of PST.CPA-PKE for a given decryption error distribution
def gen_failure_probability_compromise(error_distribution_pessimistic,error_distribution_compromise,ring,bound,key_size,B):
    if(ring!=0):
        fail_compromise = tail_probability(error_distribution_compromise, floor(bound/2.)) * 2.0    #Ring case. Failure=2 * T.P.(chi_summand1, Delta_P/2)).
        fail_compromise *= (key_size/B)
        log_fail_compromise = log(fail_compromise + 2.**(-300))/log(2)
    else:
        log_fail_compromise = gen_failure_probability_pessimistic(error_distribution_pessimistic,ring,bound,key_size,B)    #Compromise case does not apply to the Nonring case.
    #print "log_fail_compromise",log_fail_compromise
    return log_fail_compromise


#Generate the quasi-optimistic failure probability of PST.CPA-PKE for a given decryption error distribution
def gen_failure_probability_quasioptimistic(error_distribution,ring,bound,key_size,B):
    fail_quasioptimistic = tail_probability(error_distribution, floor(bound))
    fail_quasioptimistic *= (key_size/B)
    #print "log_fail_quasioptimistic",log(fail_quasioptimistic + 2.**(-300))/log(2)
    return log(fail_quasioptimistic + 2.**(-300))/log(2)        


#Generate the optimistic failure probability of PST.CPA-PKE for a given decryption error distribution
def gen_failure_probability_optimistic(error_distribution,ring,bound,key_size,B):
    fail_optimistic = tail_probability(error_distribution, floor(bound))
    fail_optimistic *= (key_size/B)
    #print "log_fail_optimistic",log(fail_optimistic + 2.**(-300))/log(2)
    return log(fail_optimistic + 2.**(-300))/log(2)     


def gen_explicit_psterr_optimistic(error_distribution_quasioptimistic,q,p,t):
    #print "\ngen_explicit_psterr_optimistic(-,q,p,t): ",q,p,t
    d=gcd(q,p)
    error_distribution_optimistic = law_convolution(error_distribution_quasioptimistic, build_mod_switching_error_law((p*q),(d*t)))
    return error_distribution_optimistic


# Explicitly calculate the failure probability of PST.PKE ring version.
# WARNING: Below analysis is specifically for the prime cyclotomic ring. See decryptionerror.tex for more details.
def gen_explicit_psterr(q,p,n,hammwt,B,ring,is_hnf=False,is_binary=False,is_ternary=False,is_sparsehnf=False,is_spbinary=False,is_spternary=True):
    #print "\ngen_explicit_psterr(q,p,n,hammwt,B,ring): ",q,p,n,hammwt,B,ring
    if(is_sparsehnf or is_spbinary or is_spternary):
        h=hammwt
    else:
        h=n
    d=gcd(q,p)
    uniform_dist_qd=build_mod_switching_error_law(q,d)    #the uniform distribution in (-q/2d,q/2d] over the integers.
    uniform_dist_pd=build_mod_switching_error_law(p,d)    #the uniform distribution in (-p/2d,p/2d] over the integers.

    # ERROR DISTRIBUTION CALCULATION PESSIMISTIC APPROACH. Read decryptionerror.tex for explanation.
    chi_E1pr = law_convolution(uniform_dist_qd,uniform_dist_pd)    #Distribution of error matrix E'_1.
    #Distribution of error matrix E_2 is the same as chi_Eipr 
    chi_E1pr_R_k1=iter_law_convolution(chi_E1pr,h)    #h-fold convolution of chi_E'_1. distribution of component of E'_1.R's polynomial products. Note that chi_E1pr_R_k2=chi_E1pr_R_k1.
    #Distribution of first, and also second component of S.E_2 is the same as chi_E1pr_R_k1.
    error_dist_pessimistic = chi_E1pr_R_k1

    # ERROR DISTRIBUTION CALCULATION COMPROMISE APPROACH. Read decryptionerror.tex for explanation.
    chi_summand1 = law_convolution(chi_E1pr_R_k1, chi_E1pr_R_k1)    #equivalent to the 2h-fold convolution of chi_E1pr, i.e., chi_E'_1.
    #chi_summand2 = chi_summand1
    error_dist_compromise = chi_summand1

    # ERROR DISTRIBUTION CALCULATION OPTIMISTIC APPROACH. Read decryptionerror.tex for explanation.
    chi_E1 = law_convolution(chi_E1pr, uniform_dist_pd)
    if(ring!=0):    #Ring case 1 or 2.
        chi_E1R_SE2 = law_convolution(iter_law_convolution(chi_E1,(2*h)), chi_summand1)    #Ring case. Equivalent to convolving the 2h-fold convolution of chi_E1, with the 2h-fold convolution of chi_E1pr (AKA chi_E'_1). Note that the latter distribution is exactly chi_summand1.
    else:
        chi_E1R_SE2 = law_convolution(iter_law_convolution(chi_E1, h), chi_E1pr_R_k1)    #Nonring case. Equivalent to convolving the h-fold convolution of chi_E1, with the h-fold convolution of chi_E'_1. Note that the latter distribution is exactly chi_E1pr_R_k1. 
    error_dist_quasioptimistic = law_convolution(chi_E1R_SE2, chi_E1pr) 

    return (error_dist_pessimistic, error_dist_compromise, error_dist_quasioptimistic)


#Summarize all 4 flavors of explicit failure probability computation
def summarize_explicit_pst_failureprob(q,p,t,n,h,B,ring,key_size,is_hnf=False,is_binary=False,is_ternary=False,is_sparsehnf=False,is_spbinary=False,is_spternary=True):
    print "q,p,t,n,h,B,ring,key_size,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary: ",q,p,t,n,h,B,ring,key_size,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary
    #Generate the decryption error distributions via the 3 (and 1 more) approaches.
    (distpess, distcomp, distquasiopt)=gen_explicit_psterr(q,p,n,h,B,ring,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary)
    distopt=gen_explicit_psterr_optimistic(distquasiopt,q,p,t)
    #Generate failure probabilities for all approaches.
    failpess=gen_failure_probability_pessimistic(distpess,ring,pst_cpapke_delta_bound_pessimistic(q,p,B,t,h),key_size,B)
    failcomp=gen_failure_probability_compromise(distpess,distcomp,ring,pst_cpapke_delta_bound_pessimistic(q,p,B,t,h),key_size,B)
    failquasiopt=gen_failure_probability_quasioptimistic(distquasiopt,ring,pst_cpapke_delta_bound_quasioptimistic(q,p,B,t,h),key_size,B)
    failopt=gen_failure_probability_optimistic(distopt,ring,pst_cpapke_delta_bound(q,p,B,t,h),key_size,B)
    return (failpess, failcomp, failquasiopt, failopt)


# Calculate the failure probability for the NTT-friendly ring case, i.e., where q is a prime, and p does not divide q, 
# both by using the Chernoff tail probability (non-tight), and by using an approximation of the PST decapsulation error (scaled) that is very close to a discrete Gaussian.
# @author: Ronald.
def pstfailure_primeq_F(x):
    return log(sinh(x)/x)


def pstfailure_primeq(q,p,t,n,h,B,ring,key_size,is_hnf=False,is_binary=False,is_ternary=False,is_sparsehnf=False,is_spbinary=False,is_spternary=True):
    twoB = 2**B
    if(ring!=1):    #should not be calling this function if not in the NTT-friendly ring instantiation.
        print "Should not be calling pstfailure_primeq() for a non-NTT friendly ring instantiation. Abort."
        exit()
    x = var('a')
    mm = find_local_minimum( (4*h+1)*pstfailure_primeq_F(x/2) + (6*h+1)*pstfailure_primeq_F(x*p/2/q) + pstfailure_primeq_F(x*p/t) - p*x/2/twoB, 0.1, 5 )     #WARNING: find_local_minimum() requires SAGE.
    pbitchernov =  log(2*key_size*exp(mm[0]),2).n()
    v = float(( (4*h) + 1 + ((6*h + 1)*((p/q)**2)) + ((p/t)**2) )/12)
    pbitgauss = log(ceil(key_size/B)*erfc( p/2/twoB/sqrt(2.0*v )), 2.0).n()        #taking the union bound over the desired keylength, not over n.
    return (pbitgauss,pbitchernov)    #return the PST failure probability (log_2) using the approximate discret Gaussian method (tighter/more accurate bound) and the Chernoff tail probability method (less accurate, more conservative).


## END operational correctness analysis  ##


#Returns the next largest prime from the list of recommended primes for the NTRU parameter n.
def next_NTRU_prime_n(n_input):
    n_ntru_list=[419, 421, 443, 461, 467, 491, 509, 523, 541, 547, 557, 563, 587, 613, 619, 653, 659, 661, 677, 701, 709, 757, 773, 787, 797, 821, 827, 829, 853, 859, 877, 883, 907, 941, 947]
    i = 0
    while (n_ntru_list[i]<=n_input and (i+1)!=len(n_ntru_list)):
        i = i + 1
    return n_ntru_list[i]


#Generate a dictionary/HashMap of the form KEY=prime q, VALUE=List of prime nplus1's satisfying q=1(mod nplus1).
def get_dict_qnplus1(nplus1_min,nplus1_max,q_min,q_max,th):
    qprimes_list=[]
    q=q_min
    while(q<=q_max):
        if(isprime(q)):
            qprimes_list.append(q)
        q=nextprime(q)
    #print "qprimes_list",qprimes_list
    dict_qm=defaultdict(list)
    for q in qprimes_list:
        nplus1=nplus1_min
        nplus1_list=[]
        while(nplus1<=nplus1_max):
            if(isprime(nplus1) and q%nplus1==1 and nplus1 not in nplus1_list and (close_to_power_of_two(q,th)==1)):
                nplus1_list.append(nplus1)
            nplus1=nextprime(nplus1)
        if nplus1_list:
            dict_qm[q]=nplus1_list
    #print "dict_qm",dict_qm
    if not dict_qm:
        print "Warning. Dictionary of (q,m[]) is empty. Check bounds for q and nplus1."
        return None
    else:
        return dict_qm


#Return the next largest prime from a list of prime nplus1's such that the nplus1-th cyclotomic polynomial \phi_nplus1(x) is irreducible modulo 2, and modulo 2**k for any k.
def next_cyclotomic_prime_nplus1_qpow2(nplus1_input):
    nplus1_list_pow2q = [2, 3, 5, 11, 13, 19, 29, 37, 53, 59, 61, 67, 83, 101, 107, 131, 139, 149, 163, 173, 179, 181, 197, 211, 227, 269, 293, 317, 347, 349, 373, 379, 389, 419, 421, 443, 461, 467, 491, 509, 523, 541, 547, 557, 563, 587, 613, 619, 653, 659, 661, 677, 701, 709, 757, 773, 787, 797, 821, 827, 829, 853, 859, 877, 883, 907, 941, 947]
    i = 0
    while (nplus1_list_pow2q[i]<=nplus1_input and (i+1)!=len(nplus1_list_pow2q)):
        i = i + 1
    return nplus1_list_pow2q[i]


#Return the next largest item in a given list. If no such item exists, return the first one.
def next_largest_item(thelist,theinput):
    i=0
    thelist=sorted(thelist)
    if thelist:    #Sanity check, make sure list is not empty
        while(thelist[i]<=theinput and (i+1)!=len(thelist)):
            i+=1
        return thelist[i]
    else:
        print "\nnext_largest_item(): thelist is empty. Abort."
        exit()
        return -1


# Read a Python dictionary stored in a text file into a dictionary data type
def load_dictionary_from_file(inpfilename):
    with open(inpfilename, 'r') as inpfile:
        diction_str = inpfile.read()
        diction = literal_eval(diction_str)
    return diction


# Lookup the failure probability for a given parameter set, calculated using an approximation to a discrete Gaussian distribution.
def lookup_primeq_fp(primeq_fp_lookuptable,q,p,t,n,hammwt,B,ring,key_size):
    if primeq_fp_lookuptable:
        key = str(q)+"_"+str(p)+"_"+str(t)+"_"+str(n)+"_"+str(hammwt)+"_"+str(B)+"_"+str(ring)+"_"+str(key_size)
        try:
            return primeq_fp_lookuptable[key]
        except KeyError:
            print "Lookup failed with key: "+str(key)+". No such parameter set exists in the F.P. lookup table."
            return log_infinity
    else:
        print "Failure probability lookup table is empty. Abort."
        return log_infinity
    #Run it
    #pst_lookuptable_filename = "/home/310220923/Software/nistpq/parameters/ntt_ringcase/Chernoff_FP/pst_fp_lookuptable.txt"
    #pst_lookuptable = load_dictionary_from_file(pst_lookuptable_filename)
    #Print the PST failure probability lookuptable
    #for key in list(sorted(pst_lookuptable.keys())):
    #    print "Parameter set: "+str(key)+"\t\tFailure rate: "+str(pst_lookuptable[key])
	

def summarize_params(file,paramlist):
    for param in paramlist:
        file.write(str("----------------------------------------------------------------------------\n")+str(param)+str("\r\n"))
    return 1


# inudsectype = 0 means its CPA secure. indsectype = 1 means its CCA secure.
# Guide to naming of schemes:
# Possible configurations: 30.
#
# Type of instantiation:
#
## Nonring:           round2_---_nr_q2_level1, ..., round2_---_nr_q2_level5    (q2 represents that q is a power of 2),
## Unified Ring:      round2_---_r_q2_level1, ..., round2_---_r_q2_level5,
## NTT-friendly Ring: round2_---_r_qp_level1, ..., round2_---_r_qp_level5   (qp represents that q is prime).
#
# Type of scheme, depending on type of targeted security::
#
## Targeting IND-CPA security -> KEM scheme:        round2_kem_--_--_level1, ..., round2_kem_--_--_level5,
## Targeting IND-CCA security -> PKE scheme:        round2_pke_--_--_level1, ..., round2_pke_--_--_level5.
#
def print_c_file(paramlist_nist, outfile_c, levels, security_level, security_buffer, indsectype, ring, log_target_failure):
    s ="/* Copyright (c) 2017 Koninklijke Philips N.V. All rights reserved. A copyright*/\n"
    s +=  "/* license for redistribution and use in source and binary forms, with or*/\n"
    s += "/* without modification, is hereby granted for non-commercial, experimental and*/\n"
    s += "/* research purposes, provided that the following conditions are met:*/\n"
    s += "/* - Redistributions of source code must retain the above copyright notice, this*/\n"
    s += "/*   list of conditions and the following disclaimer.*/\n"
    s += "/* - Redistributions in binary form must reproduce the above copyright notice,*/\n"
    s += "/*   this list of conditions and the following disclaimer in the documentation*/\n"
    s += "/*   and/or other materials provided with the distribution. If you wish to use*/\n"
    s += "/*   this software commercially, kindly contact info.licensing@philips.com to*/\n"
    s += "/*   obtain a commercial license.*/\n"
    s += "/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS \"AS IS\"*/\n"
    s += "/* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE*/\n"
    s += "/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE*/\n"
    s += "/* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE*/\n"
    s += "/* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR*/\n"
    s += "/* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF*/\n"
    s += "/* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS*/\n"
    s += "/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN*/\n"
    s += "/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)*/\n"
    s += "/* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE*/\n"
    s += "/* POSSIBILITY OF SUCH DAMAGE.*/\n"
    s += "\n\n"

    s += "/** @cond DEVELOP\n"
    s += "* @file\n"
    s += "* Provides the mappings from the NIST algorithm parameters\n"
    s += "* `CRYPTO_SECRETKEYBYTES`, `CRYPTO_PUBLICKEYBYTES`, `CRYPTO_BYTES`, and\n"
    s += "* `CRYPTO_CIPHERBYTES` to the internal PST PQ algorithm parameters.\n"
    s += "*\n"
    s += "* @author Jose Luis Torre Arce, Hayo Baan, Sauvik Bhattacharya\n"
    s += "* @endcond\n"
    s += "*/\n\n"

    currentconfig = "round2_"

    s += "/* Round2 parameter set for NIST security level "+str(security_level)+", for "
    if(indsectype==0):
        s += "CPA KEM. \n"
        currentconfig += "kem_"
    elif(indsectype==1):
        s += "CCA PKE. \n"
        currentconfig += "pke_"
    else:
        s += "(Undefined IND security type). \n"
    if(ring==0):
        s += "*  NON-RING Instantiation, \n"
        currentconfig += "nr_q2_"
    elif(ring==1):
        s += "*  NTT-friendly RING Instantiation: Prime (n+1), prime q=1 mod (n+1). \n"
        currentconfig += "r_qp_"
    elif(ring==2):
        s += "*  Unified RING Instantiation: Prime (n+1), power-of-2 q. \n"
        currentconfig += "r_q2_"
    s += "*  Target Failure Rate (log_2): "+str(log_target_failure)+"\n"
    s += "*/\n\n"

    s += "#ifndef NIST_API_TO_PARAMS_H\n"
    s += "#define NIST_API_TO_PARAMS_H\n\n"
    
    s += "#include <stdint.h>\n\n"

    s += "/* Position of the NIST API parameters */ \n"
    s += "#define NIST_SECRET        0\n"
    s += "#define NIST_PUBLIC        1\n"
    s += "#define NIST_BYTES         2\n"
    s += "#define NIST_CIPHER        3\n"
    s += "#define NIST_SHAREDSECRET  4\n\n"

    s += "/* Position of our internal parameters */ \n"

    s += "#define POS_D               5\n"
    s += "#define POS_N               6\n"
    s += "#define POS_H               7\n"
    s += "#define POS_Q               8\n"
    s += "#define POS_P_BITS          9\n"
    s += "#define POS_T_BITS         10\n"
    s += "#define POS_N_BAR          11\n"
    s += "#define POS_M_BAR          12\n"
    s += "#define POS_B              13\n\n"

    s += "#ifdef __cplusplus\n"
    s += "extern \"C\" {\n"
    s += "#endif\n\n"
    
    #s += "/* Parameters are in order (for each level): \nSK_size, PK_size, BYTES, CT_size, RANDOM_size, NIST Keylength, NIST keystrength (quantum), PST keylength, d, n, h, q, p, t, n_bar, m_bar, B. */\n\n"
    s += "static uint16_t param_matrix[][14] = {"

    #Sanity check
    if(levels>len(paramlist_nist)):
        #Not enough "Best parameter set" objects in paramlist_nist[]. Create a few dummy objects to make up the balance and to prevent ArrayIndexOutOfBounds errors.
        for i in range(levels-len(paramlist_nist)):
            dummyparam=Param(0,0,0,1,1,1,0,1,1,1,1,1,1,1,1,log_infinity,log_infinity,log_infinity,log_infinity,0,0,0,0,0,0,0,False,0)
            paramlist_nist.append(dummyparam)
            del dummyparam
    for i in range(levels):
        secret_key_size = ceil((2*paramlist_nist[i]._n_bar*paramlist_nist[i]._dim)/8)
        ssbytes = int(security_level[0][i]/8)    #length of the shared secret in bytes
        ciphertext_size = ceil( paramlist_nist[i]._m_bar*paramlist_nist[i]._dim*log2(paramlist_nist[i]._p) /8)
        ciphertext_size += ceil( (security_level[0][i]*log2(paramlist_nist[i]._t) / paramlist_nist[i]._B) /8)
        random_size = ssbytes    #size of the seed Sigma used to create A. Taken the same as the shared secret
        public_key_size = ceil(paramlist_nist[i]._n_bar*paramlist_nist[i]._dim*log2(paramlist_nist[i]._p)/8)
        public_key_size += random_size    #Size of sigma should also be included in the PK size
        public_key_size += 1        #adding 1 byte that represents the choice of the implementation of f_n(\sigma) - the permutation function for A.
        if(indsectype==0):
            enc_overheadbytes = ssbytes    #If CPAKEM, BYTES=SHAREDSECRET
        elif(indsectype==1):
            secret_key_size += public_key_size    #For INDCCA KEM construction, the secret-key includes the actual secret-key itself, but also the public-key and a dummy shared secret
            secret_key_size += ssbytes
            ciphertext_size += ssbytes        #For INDCCA KEM construction, the ciphertext/encapsulation contains another term that has length equal to the shared-secret size.
            aesgcm_authtag_bytes = 16.0    #represents the authentication tag added by AESGCM, our DEM.
            aesgcm_iv = 12.0    #representation of the AESGCM initialization vector
            ciphertext_size += aesgcm_authtag_bytes + aesgcm_iv
            enc_overheadbytes = ciphertext_size    #If CCAPKE, BYTES=Original ciphertext size + SHAREDSECRET (for the G components in the CCA ciphertext)
        s += "\n\t{"
        s += str((int)(secret_key_size))+","    #SK size. WARNING: For INDCCA KEM construction, the secret-key includes the actual secret-key itself, but also the public-key and a dummy shared secret.
        s += str((int)(public_key_size))+","    #PK size
        # Writing BYTES to config file. If CPAKEM, BYTES=SHAREDSECRET. If CCAPKE, BYTES=Original ciphertext size + SHAREDSECRET
        s += str(enc_overheadbytes)+","    #Encryption overhead in bytes, represented by BYTES
        s += str((int)(ciphertext_size))+","    #CT size
        s += str(ssbytes)+","    #shared secret length
        s += str(paramlist_nist[i]._dim)+","    #Lattice dimension
        s += str(paramlist_nist[i]._n)+","
        hamm_wt = int(floor(paramlist_nist[i]._theta * paramlist_nist[i]._dim))
        if((hamm_wt%2) != 0):
            hamm_wt -= 1    #restrict to even Hamming weights
        s += str(hamm_wt)+","    #Hamming weight
        s += str((int)(paramlist_nist[i]._q))+","    #Large modulus
        s += str((int)(log(paramlist_nist[i]._p)/log(2)))+","    #Rounding modulus
        s += str((int)(log(paramlist_nist[i]._t)/log(2)))+","    #Rounding modulus
        s += str(paramlist_nist[i]._n_bar)+","
        s += str(paramlist_nist[i]._m_bar)+","
        if(i != (levels-1)):
            s += str(paramlist_nist[i]._B)+"},"
        else:
            #For the last row containing the last parameter set that was written into the C file, do not add the last ","
            s += str(paramlist_nist[i]._B)+"}"

        currentconfig += "level_"

        s += "\t/* "+str(currentconfig)+"["
        s += str(security_level[0][i])+", "+str(security_level[1][i] - security_buffer)+"] */"
    s += "\n"
    s += "};\n\n"

    s += "#ifdef __cplusplus\n"
    s += "}\n"
    s += "#endif\n\n"
    s += "#endif /* NIST_API_TO_PARAMS_H */\n"

    outfile_c.write(s)
    return 0


def print_latexsummary_file(paramlist_nist, outfile_latex, levels, indsectype):
    s = "%% Summary of all Round2 parameters %%\n"
    s += "\documentclass{article} \n\usepackage{amsmath} \n\usepackage{booktabs}\n\n "
    s += "\\begin{document}\n "
    s += "\\begin{table*}[t]\n "
    s += "\centering\n "
    s += "\caption{Round2 Parameter Summary targeting "
    if(indsectype==0):
        s += "IND-CPA security (Round2.KEM).}\n "
    elif(indsectype==1):
        s += "IND-CCA security (Round2.PKE).}\n "
    s += "\\begin{tabular}{l|"
    for i in range(levels):
        s += "c"
    s += "}\n "
    s += "\\toprule\n "
    s += "\\textbf{Security Level} "
    for i in range(levels):
        s += "& \\textbf{NIST"+str(i+1)+"}"
    s += "\\\ \n "
    s += "\midrule\n "
    
    #Sanity check
    if(levels>len(paramlist_nist)):
        #Not enough "Best parameter set" objects in paramlist_nist[]. Create a few dummy objects to make up the balance and to prevent ArrayIndexOutOfBounds errors.
        for i in range(levels-len(paramlist_nist)):
            dummyparam=Param(0,0,0,1,1,1,0,1,1,1,1,1,1,1,1,log_infinity,log_infinity,log_infinity,log_infinity,0,0,0,0,0,0,0,False,0)
            paramlist_nist.append(dummyparam)
            del dummyparam
    #write bandwidth data for all levels.
    s_bw = "Bandwidth (B) "
    #write failure rate data for all levels.
    s_fp = "Failure rate "
    #write problem dimension data for all levels.
    s_dim = "$d$ "
    #write parameter n data for all levels.
    s_n = "$n$ "
    #write Hamming weight data for all levels.
    s_h = "$h$ "
    #write large modulus data for all levels.
    s_q = "$q$ "
    #write rounding modulus data for all levels. 
    s_p = "$p$ "
    #write ciphertext compression modulus data for all levels.
    s_t = "$t$ "
    #write message space modulus data for all levels.
    s_B = "$B$ "
    #write Primal attack cost for all levels.
    s_primal = "Security (Primal) "
    #write Dual attack cost for all levels.
    s_dual = "Security (Dual) "
    #write Hybrid attack cost for all levels.
    s_hybrid = "Security (Hybrid) "
    #write Albrecht Sparse Secret attakc cost for all levels.
    s_albrecht = "Security (Albrecht) "
    for i in range(levels):
        s_bw += "& "+str(paramlist_nist[i]._bw._bw_pstkem)
        s_fp += "& ${2}^{"+str(int(floor(paramlist_nist[i]._expl_fail_optimistic)))+"}$"
        s_dim += "& "+str(paramlist_nist[i]._dim)
        s_n += "& "+str(paramlist_nist[i]._n)
        s_h += "& $"+str(paramlist_nist[i]._theta)+"\cdot"+str(paramlist_nist[i]._dim)+"$ "
        s_q += "& "+str(paramlist_nist[i]._q)
        s_p += "& "+str(paramlist_nist[i]._p)
        s_t += "& "+str(paramlist_nist[i]._t)
        s_B += "& "+str(paramlist_nist[i]._B)
        s_primal += "& "+str(int(floor(paramlist_nist[i]._primal_best_cost)))
        s_dual += "& "+str(int(floor(paramlist_nist[i]._dual_best_cost)))
        s_hybrid += "& "+str(int(floor(paramlist_nist[i]._hybrid_best_cost)))
        s_albrecht += "& "+str(int(floor(paramlist_nist[i]._albrechtattack_cost)))
    s_bw += "\\\ \n "
    s_fp += "\\\ \n "
    s_dim += "\\\ \n "
    s_n += "\\\ \n "
    s_h += "\\\ \n "
    s_q += "\\\ \n "
    s_p += "\\\ \n "
    s_t += "\\\ \n "
    s_B += "\\\ \n "
    s_primal += "\\\ \n "
    s_dual += "\\\ \n "
    s_hybrid += "\\\ \n "
    s_albrecht += "\\\ \n "
    s = s + s_bw + s_fp + s_dim + s_n + s_h + s_q + s_p + s_t + s_B + s_primal + s_dual + s_hybrid + s_albrecht
    s += "\midrule\n "
    s += "Bandwidth (Frodo) (B) & \multicolumn{"+str(levels)+"}{c}{"+str(paramlist_nist[0]._bw._bw_frodo)+"}\\\ \n "
    s += "Failure rate (Frodo) & \multicolumn{"+str(levels)+"}{c}{${2}^{-39}$}\\\ \n "
    s += "\midrule\n "
    s += "Bandwidth (Kyber CPA) (B) & \multicolumn{"+str(levels)+"}{c}{"+str(paramlist_nist[0]._bw._bw_kybercpakem)+"}\\\ \n "
    s += "Failure (Kyber CPA) & \multicolumn{"+str(levels)+"}{c}{${2}^{-72}$} \\\ \n "
    s += "\midrule\n "
    s += "Bandwidth (Kyber CCA) (B) & \multicolumn{"+str(levels)+"}{c}{"+str(paramlist_nist[0]._bw._bw_kyberccakem)+"}\\\ \n "
    s += "Failure (Kyber CCA) & \multicolumn{"+str(levels)+"}{c}{${2}^{-142}$} \\\ \n "
    s += "\\bottomrule\n "
    s += "\end{tabular}\n "
    s += "\end{table*}\n "
    s += "\end{document}\n " 
    outfile_latex.write(s)
    return 0
    


#Obtain NIST parameters that are both operationally correct with at worst a targeted failure probability, and also secure against the (scaled) primal and dual attacks.
#Methodology: Iterate over a chosen range of values for q, chosen range of values for p, find the smallest n that is secure for this (q,p,n) combination, then check whether this parameter tuple is also operationally correct,
#by ranging over all possible combinations of B,b_h,n_bar,m_bar in order to optimize bandwidth.
def get_params_nist(is_test, indsectype, block_sizes_nonring, block_sizes_ring, security_level, security_buffer, is_ring, levels, outfile, outfile_c, outfile_latex, theta_min, theta_range, theta_step, target_failure, qmin, qmax, minroundbits_qp, minroundbits_pt, pmin, nmin, nplus1_min, nplus1_max, th, nbarmin, nbarmax, step, Attack, FP, cost, use_lwr_reduction, use_sufficient_decryptionsuccess, pst_prime_fp_lookuptable_filename):
    is_hnf=False
    is_binary=False
    is_ternary=False
    is_sparsehnf=False
    is_spbinary=False
    is_spternary=True
    #Sanity check
    if((len(block_sizes_nonring)!=levels) or (len(block_sizes_ring)!=levels)):
        print "Number of items in BKZ blocksizes array or Target failures array inconsistent with number of security levels provided. Abort."
        exit()
    if not is_test:
        outfile.write(str("###########################################################################################################################################\n\n"))
    bw_criteria = log_infinity
    for ring in is_ring:
            #Choose BKZ blocksizes depending on whether the lattice is Euclidean or Ideal.
            if(ring==0):
                block_sizes=block_sizes_nonring
            else:
                block_sizes=block_sizes_ring
            if(ring==1):
                dict_qnplus1 = get_dict_qnplus1(nplus1_min,nplus1_max,1200,2**qmax,th)
                #DEBUG---> dict_qnplus1 = get_dict_qnplus1(nplus1_min,nplus1_max,2**qmin,2**qmax,th)    #Generate the dictionary/HashMap of prime q->List of prime m's.    WARNING! qmin,qmax is actually in log_2/bits!!
                #To speed up the script, an exhaustiv list of F.P.s using the above method has been precomputed for a large range of NTT Ring PST parameters.Load this into a dictionary/hashmap of PST Parameters->F.P.
                pst_primeq_fp_lookuptable = load_dictionary_from_file(pst_prime_fp_lookuptable_filename)
	    #We wish to find the smallest q for which there exists a small enough n for which the primal/dual attack fails even if the attacker can query upto m_max samples (2n for spRING, n+n_bar for spKEX).
            if(ring==0 or ring==2):    #Nonring case where q can be anything but preferred to be power-of-2, or the Ring Case 2 where q is power-of-2 and the reduction polynomial is irreducible modulo 2.
	        log_q0=qmin                                #Starting value of modulus q is 2^10.
	        log_qn=qmax                                #Largest practical value of q.
	        q_items=int(ceil(log_qn-log_q0+1))
	        q_vals=[(2**(log_q0+i)) for i in range(q_items)]            #List of q's to evaluate for each security level.
            elif(ring==1):    #Ring case 1, q prime, nplus1 prime such that the reduction polynomial is irreducible.
                q_vals=sorted(list(dict_qnplus1.keys()))    #List of keys in the dictionary of Prime_q->Prime_m[], sorted in ascending order.
                #print "Ring case 1. q_vals is",q_vals
	    paramlist=[]                                                #List of operational, secure parameter sets. 
	    paramlist_nist=[]						#List of best parameter sets for NIST
	    bw_min = [log_infinity for _ in range(levels)]  #Store the minimum bandwidth achieved for each security level
            #Update nbarmin and nbarmax according to the ring or non-ring case to decrease search space.
            if(ring==0):
                nbarmin=5    #For non-ring cases, large nbar is needed to extract enough key bits. No point looking for a small nbar.
            else:
                nbarmax=2    #For ring cases, a very small nbar is enough to get all required key bits
	    theta_items = int(ceil(theta_range/theta_step)) + 1
	    theta_vals=[(theta_min + theta_step*i) for i in range(theta_items)]            #List of theta values to evaluate
	    for theta_idx in range(len(theta_vals)):
		    theta = theta_vals[theta_idx]
                    q_idx = -1
                    #print "About to enter q loop. len(q_vals)",len(q_vals),"and q_vals",q_vals
		    while(q_idx < len(q_vals)-1):        #for q_idx in range(len(q_vals)):    Replacing the for loop with a while to make possible explicit updating of q_idx depending on conditions inside loop body.
                        #print "Entered q loop. q_idx",q_idx
                        skip_current_iteration = False     #reset the iteration skipping flag because we are starting with a fresh q..
                        q_idx += 1    #first time, sets q_idx=0 and starts the loop.
		        #Find minimum secure n for all three security-levels, and for each of the 'q_items' AKA len(q_vals) number of q's for each level.
			q = [(q_vals[q_idx]) for _ in range(levels)]            #Check for all security levels
                        print "Progress is: " +str(100*((theta_idx*len(q_vals)+q_idx ) / (len(q_vals)*theta_items)))+ "%"+"\t\tcurrent q",q_vals[q_idx]
			nminn = nmin
			if(ring==1):
                            nminn=next_largest_item(dict_qnplus1[q_vals[q_idx]],nmin+1)-1    #Subtracting 1 since dict_qnplus1 contains (q,nplus1[]) entries.
                            #print "Ring case 1. dict_qnplus1[q_vals[q_idx]] is",dict_qnplus1[q_vals[q_idx]]," and nminn ",nminn
                        elif(ring==2):
			    nminn=next_cyclotomic_prime_nplus1_qpow2(nmin+1)-1    #Subtracting 1 since this function returns values of nplus1.
			n = [nminn for _ in range(levels)]     
			m = [1 for _ in range(levels)]          #Starting value of m. Can increase upto "max_m".
			
			#Range through a set of p's, trying to optimize over security and failure probability
			log_p_min= pmin                                     
			#the ratio between q and p should be at least 2^minroundbits_qp, i.e., at least minroundbits_qp bits of noise. 
			#IF no improved reconciliation is used, can lower this requirement since otherwise the noise is too large to be reconciled by exchanging only 1 reconciliation bit per component.
			log_p_max=int(floor(log2(q_vals[q_idx])) - minroundbits_qp)
			p_items = int(ceil(log_p_max - log_p_min + 1))
			p_vals = [(2**(log_p_min+i)) for i in range(p_items)]    #List of p's to evaluate for each security level.

                        p_idx = -1
			while(p_idx < len(p_vals)-1 and not skip_current_iteration):        #Use while loop not for.
                            #print "Entered p loop. p_idx",p_idx
                            p_idx += 1    #1st time, sets p_idx = 0 and starts loop.
			    p = [(p_vals[p_idx]) for _ in range(levels)]    #Check for all security levels.
			
			    #Initialize everything
			    primal_flags=[True for _ in range(levels)]    #Flag that forces loop to break only if a n is found that is secure against the primal attack even given m_max samples.
			    dual_flags=[True for _ in range(levels)]      #Flag that forces loop to break only if a n is found that is secure against the dual attack even given m_max samples.
			    #Flag that forces loop to break only if a n is found that is secure against the hybrid MITM and lattice reduction attack against LWR(q,p) even given m_max samples.
			    hybridattack_flags=[True for _ in range(levels)]
                            albrechtspattack_flags=[True for _ in range(levels)]
			    n_maxunsafe=[log_infinity for _ in range(levels)]
			    m_min=[log_infinity for _ in range(levels)]
			    n_minsafe=[0 for _ in range(levels)]
			    
			    # For the current q, find the smallest n satisfying the following 2 conditions: 
			    # (1) primal (embedding) attack FAILS to find the unique shortest vector of required length, (we have a nice formula to determine this condition)
			    # (2) dual cost attack is at least 2^128.
                            i = -1
			    while(i < levels-1 and not skip_current_iteration):
                                #print "Entered levels loop, i",i
                                i += 1    #start loop.
			        current_sec_level = 2        # This refers to the quantum-security level. If 1, then it is classical level. If 3, it is paranoid.
			        #Set keylength in bits according to security level.
                                n_bar = nbarmin-1
			        while(n_bar < nbarmax-1 and not skip_current_iteration):        #for n_bar in range(nbarmin,nbarmax):
                                    #print "Entered n_bar loop. n_bar",n_bar
                                    n_bar += 1
                                    m_bar = nbarmin-1
			            while(m_bar < nbarmax-1 and not skip_current_iteration):        #for m_bar in range(nbarmin,nbarmax):
                                        #print "Entered m_bar loop. m_bar",m_bar," corresponding B",get_B(n_bar,m_bar+1,security_level[0][i])
                                        m_bar += 1
					keylen = security_level[0][i]         
					if (ring == 0):
						B = get_B(n_bar,m_bar,keylen)
					else:              
				                B = 1
			                sd = float(q[i]/(2*sqrt(3)*p[i]))    #LWR rounding+scale-down error's standard deviation.
                                        #print "q list",q,"\tp list",p,"\tn list",n,"\tm list",m,"\tn_bar",n_bar,"\tm_bar",m_bar,"\tB",B
				
					    #############################################
					    #############################################
					    ######    fast  search in big steps    ######			
					    #############################################
					    #############################################
					    
                                        if((8%B)==0):    #we restrict B to be a divisor of 8 to fit implementation choices.
					    #Analyzing security of parameters against the NTRU hybrid MITM and lattice reduction attack. WARNING: sparse-ternary secret is assumed.
                                            #print "DEBUG. Start hybrid attack analysis with n",n[i],"q",q[i],"p",p[i],"n_bar,m_bar",n_bar,m_bar,"theta",theta
					    while(Attack[2] and hybridattack_flags[i] and not skip_current_iteration):
						max_samples = n[i] + max(n_bar,m_bar)    #Max. number of queriable samples is n_bar+n, or m_bar+n, whichever gives more samples.
						(hybrid_cost_mitm,hybrid_cost_lr,hybrid_r)=hybrid_attack_cost_r_nist(q[i],p[i],n[i],max_samples,theta,current_sec_level,ring,cost)
						if(hybrid_cost_lr<security_level[1][i]): 
						    #Hybrid attack succeeds. Try the next n.
						    n_maxunsafe[i]=n[i]
						    if (ring==0):
							n[i]+=step
						    elif(ring==1):
                                                        if(next_largest_item(dict_qnplus1[q[i]],n[i]+1) == (n[i]+1)):
                                                            #this means that the current list of n+1's is exhausted. There are no more suitable n+1's for this q. Proceed to the next q.
                                                            print "Skipping rest of current iteration. Breaking loop."
                                                            skip_current_iteration = True
                                                            break
							n[i]=next_largest_item(dict_qnplus1[q[i]],n[i]+1) - 1
						    elif(ring==2):
							n[i]=next_cyclotomic_prime_nplus1_qpow2(n[i]+1) - 1
                                                    #print "DEBUG. Hybrid attack succeeds. Increased n to",n[i]
						else:
						    #Update minimum safe n, against the hybrid attack
						    n_minsafe[i]=max(n_minsafe[i],n[i])
						    hybridattack_flags[i]=False
					    #print "Hybrid attack analysis done/skipped. n_minsafe",n_minsafe[i],"n[i]",n[i]," attack cost",hybrid_cost_lr," security level targeted",security_level[1][i]

					    #Analyzing security of parameters against primal embedding attack.
					    while(Attack[0] and primal_flags[i] and not skip_current_iteration):
                                                #print "DEBUG. Start primal attack analysis with n",n[i],"q",q[i],"p",p[i],"n_bar,m_bar",n_bar,m_bar,"theta",theta,"\t\tn+1_list",dict_qnplus1[q[i]]
						max_samples = n[i] + max(n_bar,m_bar)
						# Check if current parameters are secure against the primal attack
						if(primal_attack_success(sd,block_sizes[i],q[i],n[i],m[i],theta,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary)):
						    #Either or both attacks succeeded. Either n too small or m is large enough. Try larger n.
						    n_maxunsafe[i]=n[i]
						    if (ring==0):
							n[i]+=step 
						    elif(ring==1):
                                                        if(next_largest_item(dict_qnplus1[q[i]],n[i]+1) == (n[i]+1)):
                                                            #this means that the current list of n+1's is exhausted. There are no more suitable n+1's for this q. Proceed to the next q.
                                                            print "Skipping rest of current iteration. Breaking loop."
                                                            skip_current_iteration = True
                                                            break
							n[i]=next_largest_item(dict_qnplus1[q[i]],n[i]+1) - 1
						    elif(ring==2):
							n[i]=next_cyclotomic_prime_nplus1_qpow2(n[i]+1) - 1
						    m[i]=1    #reset to default
						else:
						    #Primal attack failed, also dual attack cost associated with the current n is at least 2^128. Either n too large or m too small.
						    #Try larger m.
						    m_min[i]=m[i]
						    m[i]+=1    #Search for number of samples should be done accurately
						    if(m[i] >= max_samples):
							#"Limit of m reached. No more samples can be queried for this current n. Current n safe against primal attack for all m.
							#Update minimum safe n, safe against the primal attack. It could have increased since its value was set in the dual attack analysis loop above.
							n_minsafe[i]=max(n_minsafe[i],n[i])
							m[i]=0    #reset to default
							primal_flags[i]=False
					    #print "Primal attack analysis done/skipped. n_minsafe",n_minsafe[i],"n[i]",n[i]
					    
					    #Analyzing security of parameters against dual attack.
					    attack_type=2    #dual attack
					    while(Attack[1] and dual_flags[i] and not skip_current_iteration):
                                                #print "DEBUG. Start dual attack analysis with n",n[i],"q",q[i],"p",p[i],"n_bar,m_bar",n_bar,m_bar,"theta",theta
						max_samples = n[i] + max(n_bar,m_bar)
						# Check if the current parameters are secure against the dual attack
						(dual_b,dual_m,dual_cost,dual_attack_gap)=optimizeattack_b_m(q[i],p[i],n[i],max_samples,theta,current_sec_level,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary,attack_type,ring,cost,use_lwr_reduction)
						if(dual_cost<security_level[1][i]):
						    n_maxunsafe[i]=n[i]    #min(n_maxunsafe[i],n[i])
						    if (ring==0):
							n[i]+=step 
						    elif(ring==1):
                                                        if(next_largest_item(dict_qnplus1[q[i]],n[i]+1) == (n[i]+1)):
                                                            #this means that the current list of n+1's is exhausted. There are no more suitable n+1's for this q. Proceed to the next q.
                                                            print "Skipping rest of current iteration. Breaking loop."
                                                            skip_current_iteration = True
                                                            break
							n[i]=next_largest_item(dict_qnplus1[q[i]],n[i]+1) - 1
						    elif(ring==2):
							n[i]=next_cyclotomic_prime_nplus1_qpow2(n[i]+1) - 1
						else:
						    # Update minimum safe n, safe against the Dual attack
						    n_minsafe[i]=max(n_minsafe[i],n[i])
						    m[i]=1    #reset to default
						    dual_flags[i]=False
					    #print "Dual attack analysis done/skipped. n_minsafe",n_minsafe[i],"n[i]",n[i]    #"dual cost",dual_cost,"security level targeted",security_level[1][i]
					    
					    #Analyzing security of parameters against Albrecht's Sparse Secret attack.
					    while(Attack[3] and albrechtspattack_flags[i] and not skip_current_iteration):
                                                #print "DEBUG. Start Albrecht attack analysis with n",n[i],"q",q[i],"p",p[i],"n_bar,m_bar",n_bar,m_bar,"theta",theta
						max_samples = n[i] + max(n_bar,m_bar)
						# Check if the current parameters are secure against the dual attack
						(albsp_cost,colstodrop_opt)=optimize_albrecht_spattack(q[i],p[i],n[i],max_samples,theta,current_sec_level,ring,cost,use_lwr_reduction,is_hnf,is_binary,is_ternary,is_sparsehnf,is_spbinary,is_spternary)
						if(albsp_cost<security_level[1][i]):
                                                    #print "albsp_cost<security_level[1][i]. albsp_cost",albsp_cost," and security_level[1][i]",security_level[1][i]
						    n_maxunsafe[i]=n[i]    #min(n_maxunsafe[i],n[i])
						    if (ring==0):
							n[i]+=step 
						    elif(ring==1):
                                                        print next_largest_item(dict_qnplus1[q[i]],n[i]+1), n[i]+1
                                                        if(next_largest_item(dict_qnplus1[q[i]],n[i]+1) == (n[i]+1)):
                                                            #this means that the current list of n+1's is exhausted. There are no more suitable n+1's for this q. Proceed to the next q.
                                                            print "Skipping rest of current iteration. Breaking loop."
                                                            skip_current_iteration = True
                                                            break
							n[i]=next_largest_item(dict_qnplus1[q[i]],n[i]+1) - 1
						    elif(ring==2):
							n[i]=next_cyclotomic_prime_nplus1_qpow2(n[i]+1) - 1
						else:
						    #Update minimum safe n, safe against Albrecht's sparse secret attack
						    n_minsafe[i]=max(n_minsafe[i],n[i])
						    m[i]=1    #reset to default
						    albrechtspattack_flags[i]=False
					    #print "Albrecht Sparse Secret attack analysis done/skipped. n_minsafe",n_minsafe[i],"n[i]",n[i]    #"Albrecht attack cost",albsp_cost,"security level targeted",security_level[1][i]

                                            if not skip_current_iteration:
						#Compute the key-size
						key_size = B * n_bar * m_bar
						if(ring==1 or ring==2):
						    key_size *= n_minsafe[i]
						#For the current q,p,n scan the possible values for t to find the best combination of Bandwidth and Failure rate.
						t_min = 2**B    #results in best bandwidth and worst failure probability.
						t_max = 2**(floor( log2(p[i])-minroundbits_pt ))    #results in worst bandwidth and best failure probability.
						h=int(floor(theta*n_minsafe[i]))
                                                if((h%2)!=0):        #if Hamming weight is not even, take the next smaller even Hamming weight. Optimizes implementation
                                                    h -= 1
						#Evaluate the best possible bandwidth for the above range of t's and see whether it satisfies the minimum bandwidth criteria.
						best_bw_obj = get_bw(1,q[i],p[i],n_minsafe[i],t_min,n_bar,m_bar,keylen,B,indsectype)    #1st param=1,i.e.,Post quantum seclevel.
						#print "Create best_bw_obj. Frodo BW is",best_bw_obj._bw_frodo," Kyber CPA BW is",best_bw_obj._bw_kybercpakem," Kyber CCA BW is",best_bw_obj._bw_kyberccakem
						if(indsectype == 0):    #We are considering CPA schemes
						    bw_kyber = best_bw_obj._bw_kybercpakem
						elif(indsectype == 1):    #We are considering CCA schemes
						    bw_kyber = best_bw_obj._bw_kyberccakem
						if(ring==0 and best_bw_obj._bw_frodo<bw_criteria):
						    bw_criteria = best_bw_obj._bw_frodo
						elif((ring==1 or ring==2) and bw_kyber<bw_criteria):
						    bw_criteria = bw_kyber
						if(keylen > 256 or ring==1):
                                                    #if NIST is asking for a key-length greater than 256bits, we have no baseline/existing schemes to compare against.
                                                    #The 2nd condition: for NTT Ring PST, q's and n's are typically huge. Setting a strict BW requirement results in no parameters.
						    bw_criteria = log_infinity
						#print "Checking bandwidth criteria. bw_criteria",bw_criteria,"bw_min[i]",bw_min[i],"best_bw_obj._bw_pstkem",best_bw_obj._bw_pstkem
						if(best_bw_obj._bw_pstkem<=bw_criteria and best_bw_obj._bw_pstkem<=bw_min[i]):    #Extra strict check. Consider only parameter range for which the best possible BW is less than the best BW so far.
						    #print "Passed BW criteria. t_max ",t_max,"    t_min",t_min,"    bw_criteria",bw_criteria
						    psterr_pessimistic_worst = -1 * log_infinity
						    psterr_compromise_worst = -1 * log_infinity
						    psterr_quasioptimistic_worst = -1 * log_infinity    #setting these 3 values to force a search over the entire range of t if use_sufficient_decryptionsuccess flag is TRUE, or if NTT-friendly ring is being used..
						    if not use_sufficient_decryptionsuccess and ring!=1:
							#Use explicit convolutions to compute F.P. only if BOTH (1) sufficient decryption success condition is not being used and, (2) NTT-friendly ring is not being used.
							log_current_t = ceil(log2(t_min))
							while(get_bw(1,q[i],p[i],n_minsafe[i],2**log_current_t,n_bar,m_bar,keylen,B,indsectype)._bw_pstkem<bw_criteria and 2**log_current_t<=t_max):
							    #we want to find the worst possible Failure rate for this range of t. 
							    #Instead of calculating directly for the largest possible t(=p), we calculate it for the largest t for which the BW still remains competitive.
							    log_current_t+=1
							if(2**log_current_t > t_max):
							    current_t=t_max    #no sense in considering t's that are bigger than the upper bound anyway.
							else:
							    current_t=2**log_current_t
							#print "log_current_t",log_current_t,"t_max",t_max,"current_t",current_t,"    About to check failure criteria.for q",q[i],"p",p[i],"B",B,"h",h,"keylen",keylen
							(psterr_dist_pessimistic,psterr_dist_compromise,psterr_dist_quasioptimistic) = gen_explicit_psterr(q[i],p[i],n_minsafe[i],h,B,ring)    #Returns a 3-tuple.
							psterr_pessimistic_worst=1
							psterr_compromise_worst=1
							psterr_quasioptimistic_worst=1
							psterr_optimistic_worst=1
							if(FP[0]):
							    psterr_pessimistic_worst = gen_failure_probability_pessimistic(psterr_dist_pessimistic,ring,pst_cpapke_delta_bound_pessimistic(q[i],p[i],B,current_t,h),keylen,B)
							if(FP[1]):
							    psterr_compromise_worst = gen_failure_probability_compromise(psterr_dist_pessimistic,psterr_dist_compromise,ring,pst_cpapke_delta_bound_pessimistic(q[i],p[i],B,current_t,h),keylen,B)
							if(FP[2]):
							    psterr_quasioptimistic_worst = gen_failure_probability_quasioptimistic(psterr_dist_quasioptimistic,ring,pst_cpapke_delta_bound_quasioptimistic(q[i],p[i],B,current_t,h),keylen,B)

						    #print "psterr_pessimistic_worst",psterr_pessimistic_worst,"psterr_compromise_worst",psterr_compromise_worst,"psterr_quasioptimistic_worst",psterr_quasioptimistic_worst
						    if(min(psterr_pessimistic_worst,psterr_compromise_worst,psterr_quasioptimistic_worst) < target_failure):    #to decrease the search space for t, we updated the upper bound so that bandwidth is still good and for which the worst possible failure satisfies the bare minimum criteria and can be further improved.
							if not use_sufficient_decryptionsuccess and ring!=1:
							    t_max=current_t
							for log2_t in range( int(ceil(log2(t_max))), int(ceil(log2(t_min)))-1, -1):
							    t=2**log2_t
							    #print "Actual search for t. Evaluate all 3 failure rate flavors. log2_t",log2_t,"t_max",t_max,"t_min",t_min,"q",q[i],"p",p[i],"B",B,"h",h,"keylen",keylen
							    if not use_sufficient_decryptionsuccess:
								if ring!=1:
								    #print "Using explicit convolutions to upper bound F.P. Ring type",ring
								    psterr_pessimistic_bckup = gen_failure_probability_pessimistic(psterr_dist_pessimistic,ring,pst_cpapke_delta_bound_pessimistic(q[i],p[i],B,t,h),keylen,B)
								    psterr_compromise_bckup = gen_failure_probability_compromise(psterr_dist_pessimistic,psterr_dist_compromise,ring,pst_cpapke_delta_bound_pessimistic(q[i],p[i],B,t,h),keylen,B)
								    psterr_quasioptimistic_bckup = gen_failure_probability_quasioptimistic(psterr_dist_quasioptimistic,ring,pst_cpapke_delta_bound_quasioptimistic(q[i],p[i],B,t,h),keylen,B)
								    psterr_dist_optimistic = gen_explicit_psterr_optimistic(psterr_dist_quasioptimistic,q[i],p[i],t)
								    psterr_optimistic_bckup = gen_failure_probability_optimistic(psterr_dist_optimistic,ring,pst_cpapke_delta_bound(q[i],p[i],B,t,h),keylen,B)
								    #print "psterr_pessimistic_bckup", psterr_pessimistic_bckup,"psterr_compromise_bckup",psterr_compromise_bckup,"psterr_quasioptimistic_bckup",psterr_quasioptimistic_bckup," psterr_optimistic_bckup", psterr_optimistic_bckup
								    psterr_pessimistic=1
								    psterr_compromise=1
								    psterr_quasioptimistic=1
								    psterr_optimistic=1
								    if(FP[0]):
									psterr_pessimistic = psterr_pessimistic_bckup
								    if(FP[1]):
									psterr_compromise = psterr_compromise_bckup
								    if(FP[2]):
									psterr_quasioptimistic = psterr_quasioptimistic_bckup
								    if(FP[3]):
									psterr_optimistic = psterr_optimistic_bckup
								elif ring==1:
								    #Use the approximation to a discrete Gaussian to upper bound the F.P. for the NTT-friendly ring case.
                                                                    # Look up the NTT Ring PST failure probability from the pre-computed F.P. lookup table.t 
                                                                    #print "Using the approximation to a discrete Gaussian to upper bound F.P. Ring type",ring
                                                                    psterr_primeq = lookup_primeq_fp(pst_primeq_fp_lookuptable,q[i],p[i],t,n_minsafe[i],h,B,ring,keylen)
                                                                    #if(psterr_primeq<target_failure):    #DEBUG
                                                                    #    print "psterr_primeq",psterr_primeq    #DEBUG
								    psterr_pessimistic_bckup = 1
								    psterr_compromise_bckup = 1
								    psterr_quasioptimistic_bckup = 1
								    psterr_optimistic_bckup = psterr_primeq
	 
								    psterr_pessimistic = psterr_pessimistic_bckup
								    psterr_compromise = psterr_compromise_bckup
								    psterr_quasioptimistic = psterr_quasioptimistic_bckup
								    psterr_optimistic = psterr_optimistic_bckup                                                            
							    elif use_sufficient_decryptionsuccess:
								#Extremely conservative and strict failure rate evaluation. Basically, there is no failure rate.
								#We evaluate the condition that is sufficient to ensure decryption ALWAYS succeeds.
								#print "Using the sufficient condition for PST decryption success. Ring type",ring
								if pst_sufficient_success(q[i],p[i],B,t,h,ring):
								    print "Parameters:q,p,B,t,h",q[i],p[i],B,t,h,"\tpst_sufficient_success returned",pst_sufficient_success(q[i],p[i],B,t,h,ring)
								    psterr_pessimistic_bckup = -1*log_infinity
								    psterr_compromise_bckup = -1*log_infinity
								    psterr_quasioptimistic_bckup = -1*log_infinity
								    psterr_optimistic_bckup = -1*log_infinity
								else:
								    psterr_pessimistic_bckup = 1
								    psterr_compromise_bckup = 1
								    psterr_quasioptimistic_bckup = 1
								    psterr_optimistic_bckup = 1
								psterr_pessimistic = psterr_pessimistic_bckup
								psterr_compromise = psterr_compromise_bckup
								psterr_quasioptimistic = psterr_quasioptimistic_bckup
								psterr_optimistic = psterr_optimistic_bckup
								#Finish evaluation of PST sufficient decryption-success condition

							    #Create a new potential parameter set.
							    new_param=Param(1,theta, ring,q[i],p[i],sd,n_maxunsafe[i],n_minsafe[i],keylen,security_level[1][i],key_size,t,B,n_bar,m_bar,psterr_pessimistic_bckup,psterr_compromise_bckup,psterr_quasioptimistic_bckup,psterr_optimistic_bckup,dual_cost,dual_b,dual_m,dual_attack_gap,hybrid_cost_mitm,hybrid_cost_lr,hybrid_r,use_sufficient_decryptionsuccess,indsectype)    #note that the first parameter '1' just means that it is always Post-quantum security level.
							    if(ring==0):
								current_bw_criteria = new_param._bw._bw_frodo
							    else:
								current_bw_criteria = bw_kyber    #Can reuse the same Kyber bandwidth obj from the older best_bw_obj object (see above). Won't have changed.
							    if(keylen > 256 or ring==1):    
                                                                #if NIST is asking for a key-length greater than 256bits, we have no baseline/existing schemes to compare against.
                                                                #The 2nd condition: for NTT Ring PST, q's and n's are typically huge. Setting a strict BW requirement results in no parameters.
								current_bw_criteria = log_infinity
							    #Double check everything.
							    if( (min(psterr_pessimistic,psterr_compromise,psterr_quasioptimistic,psterr_optimistic) < target_failure) and (new_param._bw._bw_pstkem<=current_bw_criteria) and (new_param._keylength>=keylen) ):
								#Recalculate primal, dual, hybrid, Albrecht attack security levels since n might have been updated.
								(new_param._primal_best_b, new_param._primal_best_m, new_param._primal_best_cost, new_param._dual_attack_gap) = optimizeattack_b_m(q[i], p[i], n_minsafe[i], n_minsafe[i] + max(n_bar,m_bar), theta, 2, 0, 0, 0, 0, 0, 1, 1,ring,cost,use_lwr_reduction)
								(new_param._dual_best_b, new_param._dual_best_m, new_param._dual_best_cost, new_param._dual_attack_gap) = optimizeattack_b_m(q[i], p[i], n_minsafe[i], n_minsafe[i] + max(n_bar,m_bar), theta, 2, 0, 0, 0, 0, 0, 1, 2,ring,cost,use_lwr_reduction)
								(new_param._hybrid_mitmcost,new_param._hybrid_bestcost_lr,new_param._hybrid_r)=hybrid_attack_cost_r_nist(q[i],p[i],n_minsafe[i],n_minsafe[i] + max(n_bar,m_bar),theta,2,ring,cost)
								(new_param._albrechtattack_cost, new_param._albrechtattack_k_opt)=optimize_albrecht_spattack(q[i],p[i],n_minsafe[i],n_minsafe[i] + max(n_bar,m_bar),theta,2,ring,cost,0,False,False,False,False,False,True)
                                                                #Check the security levels for LWR(p,t) problem.
                                                                (dummy1, dummy2, new_param._primal_best_cost_pt, dummy3) = optimizeattack_b_m(p[i], t, n_minsafe[i], n_minsafe[i] + max(n_bar,m_bar), theta, 2, 0, 0, 0, 0, 0, 1, 1,ring,cost,use_lwr_reduction)
                                                                (dummy1, dummy2, new_param._dual_best_cost_pt, dummy3) = optimizeattack_b_m(p[i], t, n_minsafe[i], n_minsafe[i] + max(n_bar,m_bar), theta, 2, 0, 0, 0, 0, 0, 1, 2,ring,cost,use_lwr_reduction)
                                                                (dummy1,new_param._hybrid_bestcost_lr_pt,dummy2)=hybrid_attack_cost_r_nist(p[i],t,n_minsafe[i],n_minsafe[i] + max(n_bar,m_bar),theta,2,ring,cost)
                                                                (new_param._albrechtattack_cost_pt, dummy1)=optimize_albrecht_spattack(p[i],t,n_minsafe[i],n_minsafe[i] + max(n_bar,m_bar),theta,2,ring,cost,0,False,False,False,False,False,True)
                                                                if(min(new_param._primal_best_cost_pt,new_param._dual_best_cost_pt,new_param._hybrid_bestcost_lr_pt,new_param._albrechtattack_cost_pt) >= security_level[1][i]):
								    paramlist.append(new_param)
								    print "Added a new parameter set!"
								    if(new_param._bw._bw_pstkem < bw_min[i]):
								        if (bw_min[i] == log_infinity):
									    paramlist_nist.append(new_param) # append new parameter first time
								        else:
									    paramlist_nist[i] = new_param
								        bw_min[i] = new_param._bw._bw_pstkem    #Update the minimum bandwidth achieved by possible spKEX parameters for current security level.
    							        else:
								    del new_param
						    #else:
							#print "Failure criteria failed for n",n_minsafe[i],"q",q[i],"p",p[i],"B",B,"n_bar",n_bar,"m_bar",m_bar,"n_bar*m_bar*B",(B*n_bar*m_bar),"current_t",current_t,"h",h,"keylen",keylen,"\tMinimum failure is",min(psterr_pessimistic_worst,psterr_compromise_worst,psterr_quasioptimistic_worst)
						else:
						    #print "Bandwidth criteria failed for n",n_minsafe[i],"q",q[i],"p",p[i],"B",B,"n_bar",n_bar,"m_bar",m_bar,"n_bar*m_bar*B",(B*n_bar*m_bar),"h",h,"keylen",keylen,"Skipping to next parameter."
						    del best_bw_obj
    if not is_test:
        outfile.write(str("#############################################################################\r\n\n")) 
        for i in range(levels):
            outfile.write("Security level for key length/strength "+str(security_level[0][i])+"/" + str(security_level[1][i])+ ": Minimum bandwidth achieved: "+str(bw_min[i])+"\r\n")
        outfile.write(str("\n#############################################################################\r\n"))    
        outfile.write(str("### Best parameters for NIST 	              #################################################\r\n"))
        outfile.write(str("#############################################################################\r\n"))
        summarize_params(outfile,paramlist_nist)
        outfile.write(str("#############################################################################\r\n"))
        outfile.write(str("### Summary of all parameter sets	              #################################################\r\n"))
        summarize_params(outfile,paramlist)
        outfile.write(str("#############################################################################\r\n"))
        print_c_file(paramlist_nist, outfile_c, levels, security_level, security_buffer, indsectype, ring, target_failure)
        print_latexsummary_file(paramlist_nist, outfile_latex, levels, indsectype)
    return paramlist


#--------------------------------------------------------------------------------------------------------------------------
#----------------------------------------------------    TEST FRAMEWORK ---------------------------------------------------
#Test the parameter search process by re-obtaining expected parameters for spKEX. Reference: Table 2, https://eprint.iacr.org/2017/709.pdf
do_test=False    #flag determining whether testing is done or no. Set to false to skip tests.
if do_test:
    is_test=True
    test_indsectype=0    #does not apply to spKEX
    test_seclevel=[[256],[128]]
    test_attack=[1,1,1]
    test_failuremodel=[1,1,1,1]
    test_cost=0.265
    test_targetfailure=-40    #in log_2
    test_is_ring=[0]    #spKEX is non ring
    test_thetamin=0.2
    test_thetarange=0
    test_thetastep=0.1
    test_minlwrbits=2
    test_nmin=700
    test_qmin=17
    test_qmax=17
    test_pmin=11
    test_nplus1_min=test_nmin 
    test_nplus1_max=2000
    test_th=256
    test_nbarmin=5
    test_nbarmax=9
    test_step=1
    test_uselwrreduction=0
    test_bkz_block_sizes=[50 for _ in range(len(test_seclevel[0]))]
    for i in range (len(test_seclevel[0])):
        while(svp_quantum_nist(test_bkz_block_sizes[i],0,test_cost)<=test_seclevel[1][i]):
            test_bkz_block_sizes[i]+=1
    print "Running tests, getting spKEX parameters...\n"
    spkex_params=get_params_nist(is_test,test_indsectype,test_bkz_block_sizes,test_bkz_block_sizes, test_seclevel, 0, test_is_ring, len(test_seclevel[0]), None, None, None, test_thetamin, test_thetarange, test_thetastep, test_targetfailure, test_qmin, test_qmax, test_minlwrbits, test_pmin, test_nmin, test_nplus1_min, test_nplus1_max, test_th, test_nbarmin, test_nbarmax, test_step, test_attack, test_failuremodel, test_cost, 0, False, None)    #using the PST sufficient decryption-success condition does not make sense with spKEX
    testok=True
    for spkex_param in spkex_params:
        if spkex_param._dim==738:
            if((round(spkex_param._primal_best_cost)!=136) or (spkex_param._primal_best_b!=479) or (round(spkex_param._hybrid_best_cost)!=128) or (round(spkex_param._hybrid_r)!=278)):
                print "Conditions failed:",round(spkex_param._primal_best_cost),spkex_param._primal_best_b,round(spkex_param._hybrid_best_cost),round(spkex_param._hybrid_r)
                testok=False
        else:
            testok=False    #No parameter was found with n_minsafe = 738. This is unexpected.
    for spkex_param in spkex_params:
        print "\n"+str(spkex_param)
    if not testok or (len(spkex_params)==0):
        print "\nTests failed. Check spKEX test parameters obtained. The required conditions:\nif((round(spkex_param._primal_best_cost)!=136) or (spkex_param._primal_best_b!=479) or (round(spkex_param._hybrid_best_cost)!=128) or (spkex_param._hybrid_r!=278))."
        exit()
    else:
        print "\nspKEX tests passed. Starting actual parameter search for PST.\n"



#---------------------------------------------------------------------------------------------------------------------------------------
#----------------------------------------------------    Get parameters for Round2   ---------------------------------------------------
#---------------------------------------------------------------------------------------------------------------------------------------



#Targeted security levels for NIST
# first array is about key length
# second array is about quantum security level
security_level = [[128, 256, 192, 256, 384], [64, 86, 96, 128, 128]]
secbuffer = 10    #represents a security buffer in bits. 
                  #Some breathing room above and beyond a given security level.
for i in range(len(security_level[1])):
    security_level[1][i] += secbuffer
print(security_level)
#configuration of script. In only executes the '1'
Attack=[1,1,1,1] # attacks to consider: primal, dual, hybrid, Albrecht's sparse secret LWE attack.
FP=[1,1,1,1] # way of obtaining FP: Explicit pessimistic/safest, 
             # Explicit compromise/fairly pessimistic, 
             # Explicit quasi-optimistic, 
             # Explicit optimistic/realistic. See function definition comments for more explanation.

use_sufficient_decryptionsuccess = False    #If this is true, then failure probabilities are not 
                                            #calculated at all, (irrespective of the approach), 
                                            #but a sufficient condition on decryption success/failure 
                                            #is used to decide whether decryption succeeds or not.
#Choose a cost
costs=[0.292,0.265,0.2075]        # [Classical, Quantum, Paranoid]
cost = costs[1]

#Flag determinining whether the type of IND security targeted is in 
#the Chosen Plaintext Attacker model, 
#or Chosen Ciphertext Attacker model. 0: CPA, 1: CCA.
indsec_types=[0,1]
#target failure rate (log2) for the INDCPA security level.
indcpa_fp = -64
#target failure rate (log2) for the INDCPA security level, 
#should be negligible in the keyspace assuming at most 
#2^64 decryption oracle queries.
indcca_fp = -128
for indsec_type in indsec_types:
    #target failure probability, depends on IND security targeted.
    if(indsec_type == 0):    #IND-CPA security targeted.
	log_target_failure = indcpa_fp
    elif(indsec_type == 1):    #IND-CCA security targeted.
	log_target_failure = indcca_fp

    # Cases: 
    # 0=Non-ring case; 
    # 1=Ring case over the m-th prime cyclotomic polynomial defined for prime m, 
    # prime modulus q satisfying q=1(mod m); 
    # 2=Ring case over the m-th prime cyclotomic polynomial defined for prime m, 
    # power-of-2 modulus q that is irreducible modulo 2 and modulo 2**k for any k.
    is_ring = [0,1,2]

    #Using the Search-LWE to Decision-LWR reduction or not. 
    # Warning: Cannot be combined with the ring case since no efficient 
    # reduction from (R)LWE to Decision-RLWR is known.
    use_lwr_reduction = 0    #If this is 1, turn ring to 0.

    #Parameters
    theta_min = 0.12     # minimum hamming_weight
    theta_range = 0.10    # range of hamming_weight to explore
    theta_step = 0.01    # step when exploring hamming_weight

    roundbits_qp = 2    #minimum desired noise for the LWR(q,p) problem, 
                        #represented as number of bits dropped during rounding, 
                        #minimum value of log2(q)-log2(p).
    roundbits_pt = 1    #minimum desired noise for the LWR(p,t) problem, 
                        #represented as number of bits dropped during rounding, 
                        #minimum value of log2(q)-log2(t).
    nmin = 350

    nplus1_min = nmin+1
    nplus1_max = 1500    #something ridiculously large

    th = log_infinity    #For ring case 1, th represents how much a prime q can 
                         #differ from the closest power-of-2 in order to be suitable 
                         #for a parameter set. Set very large to ignore this restriction.
    qmin = 8
    qmax = 12

    pmin = qmin - roundbits_qp

    nbarmin = 1    #Will be set to 5 for the non-ring case, inside get_params_nist()
    nbarmax = 11    #Will be set to 2 for the ring case, inside get_params_nist()

    step = 1 #step to scan for parameters

    pst_primeq_fp_lookuptable_filename = "/home/310220923/Software/nistpq/parameters/ntt_ringcase/pst_fp_lookuptable.txt"        #WARNING! This is hardcoded, be very careful!

    #Create a separate directory for writing all results into 
    path_name=subprocess.check_output(['pwd']).strip()        #<---- WARNING!! When running on cluster, replace with explicit path to wherever you want to write results.
    path_name=path_name.strip()
    abs_log_target_failure = abs(log_target_failure)
    dir_name="NIST_Parameters__Ring"+str(is_ring)+"_FP"+str(abs_log_target_failure)
    dir_name=path_name+str("/")+dir_name
    dir_name=dir_name.replace(" ","_")    #Remove all spaces and problematic characters
    dir_name=dir_name.replace(".","_")
    dir_name=dir_name.replace(":","_")
    dir_name=dir_name.replace("-","_")
    if (False == os.path.isdir(dir_name)):
	cmd="mkdir "+dir_name
	process=subprocess.Popen(cmd, shell=True)
	process.wait()
	process.wait()
	ss = "Python subprocess return code:" + str(process.returncode)
	print ss
    s=time.asctime()
    
    outfilename=dir_name+"/"
    outfilename=outfilename+"_Complete report_"
    outfilename=outfilename + "_"
    outfilename=outfilename + "SecurityLevel_" + str(security_level) + "_FP"+str(abs_log_target_failure)+"__" 
    outfilename=outfilename + str(s) 
    outfilename=outfilename.replace(" ","_")    #Remove all spaces and problematic characters.
    outfilename=outfilename.replace(".","_")
    outfilename=outfilename.replace(":","_")
    outfilename=outfilename.replace("-","_")
    outfilename=outfilename +".txt"
    outfile=open(outfilename,'w')
    paramfilename=outfilename
    outfile.seek(0)                #Start writing at the beginning of the file
    
    outfilename=dir_name+"/"
    outfilename=outfilename +"_nist_api_to_algorithm_parameters_"
    outfilename=outfilename + "_"
    outfilename=outfilename + "SecurityLevel_" + str(security_level) + "_FP"+str(abs_log_target_failure)+"__"
    outfilename=outfilename + str(s)
    outfilename=outfilename.replace(" ","_") 
    outfilename=outfilename.replace(".","_")
    outfilename=outfilename.replace(":","_")
    outfilename=outfilename.replace("-","_")
    outfilename=outfilename +".h"
    outfile_c=open(outfilename,'ww')
    apifilename=outfilename
    outfile_c.seek(0)
    
    outfilename=dir_name+"/"
    outfilename=outfilename +"_PST_parameter_summary_"
    outfilename=outfilename + "_"
    outfilename=outfilename + "SecurityLevel_" + str(security_level) + "_FP"+str(abs_log_target_failure)+"__"
    outfilename=outfilename + str(s)
    outfilename=outfilename.replace(" ","_") 
    outfilename=outfilename.replace(".","_")
    outfilename=outfilename.replace(":","_")
    outfilename=outfilename.replace("-","_")
    outfilename=outfilename +".tex"
    outfile_latex=open(outfilename,'ww')
    latexfilename=outfilename
    outfile_latex.seek(0)

    s="Attacks considered are " + str(Attack) + "\n"
    s += "Failure probability method is " + str(FP) + "\n"
    s += "Cost for BKZ is " + str(cost) + "\n"
    s += "Probability of failure of final key-agreement is less than 2^("+str(log_target_failure)+").\r\n"
    outfile.write(s)

    #Find BKZ block-sizes satisfying quantum security levels 
    #required by NIST when considering BKZ-b.
    #Cost of BKZ lattice-reduction is NOT ONLY a function of b, 
    #but also whether the lattice is ideal or Euclidean.
    ring=0    #nonring case
    bkz_block_sizes_nonring=[50 for _ in range(len(security_level[0]))]
    for i in range (len(security_level[0])):
	while(svp_quantum_nist(bkz_block_sizes_nonring[i],ring,cost)<=security_level[1][i]):
	    bkz_block_sizes_nonring[i]+=1
	s = "Block size for security level (Euclidean lattice)" + str(security_level[1][i]) + " is " + str(bkz_block_sizes_nonring[i]) + "\n"
	outfile.write(s)
    ring=1    #ring case
    bkz_block_sizes_ring=[50 for _ in range(len(security_level[0]))]
    for i in range (len(security_level[0])):
	while(svp_quantum_nist(bkz_block_sizes_ring[i],ring,cost)<=security_level[1][i]):
	    bkz_block_sizes_ring[i]+=1
	s = "Block size for security level (Ideal lattice)" + str(security_level[1][i]) + " is " + str(bkz_block_sizes_ring[i]) + "\n"
	outfile.write(s)
    #Sanity check
    if((len(bkz_block_sizes_nonring)!=len(security_level[0])) or (len(bkz_block_sizes_ring)!=len(security_level[0]))):
	print "Dimension of bkz_block_sizes_list is not the same as security_level[0]. Abort."
	exit()
    if(len(security_level[0]) != len(security_level[1])):
	print "Number of keylengths in security_level != Number of security levels in security_level. Abort."
	exit()

    #Get Round2 parameters
    get_params_nist(False,indsec_type,bkz_block_sizes_nonring,bkz_block_sizes_ring, security_level, secbuffer, is_ring, len(security_level[0]), outfile, outfile_c, outfile_latex, theta_min, theta_range, theta_step, log_target_failure, qmin, qmax, roundbits_qp, roundbits_pt, pmin, nmin, nplus1_min, nplus1_max, th, nbarmin, nbarmax, step, Attack, FP, cost, use_lwr_reduction, use_sufficient_decryptionsuccess, pst_primeq_fp_lookuptable_filename)
    #Close the output file
    outfile.close()
    outfile_c.close()
    outfile_latex.close()
    print "\nFinished. \nParameters file written to",str(paramfilename),"\nAPI file written to",str(apifilename),"\nLaTeX parameter summary file written to",str(latexfilename)

    #Create a PDF file using the TeX parameter summary file that was created.
    pdfcmd = "pdflatex -halt-on-error -output-directory "+str(dir_name)+" "+str(latexfilename)
    #subprocess.check_call(['pdflatex', latexfilename])
    pdfprocess=subprocess.Popen(pdfcmd, shell=True)
    pdfprocess.wait()

    #Clean up the .aux and .log files created by pdflatex 
    #as part of the PDF creation process from the TeX file.
    auxfilename = latexfilename.rsplit( ".", 1 )[ 0 ]+str(".aux")
    logfilename = latexfilename.rsplit( ".", 1 )[ 0 ]+str(".log")
    cleancmd1="rm "+str(auxfilename)
    cleanprocess1=subprocess.Popen(cleancmd1, shell=True)
    cleanprocess1.wait()
    cleancmd2="rm "+str(logfilename)
    cleanprocess2=subprocess.Popen(cleancmd2, shell=True)
    cleanprocess2.wait()
