#!/usr/bin/env sage
#from sage.all import *
from sympy import isprime,nextprime
from collections import defaultdict
log_infinity = 999999999


####---- A standalone file that creates a lookup table of failure rates for NTT-Ring PST parameter sets in a given range, and saves it as a Python dictionary to file. ----####


#Calculate log base 2 of an input
def log2(n):
    res=log(n, 2.0).n()
    return res


# Calculate the failure probability for the NTT-friendly ring case, i.e., where q is a prime, and p does not divide q, 
# both by using the Chernoff tail probability (non-tight), and by using an approximation of the PST decapsulation error (scaled) that is very close to a discrete Gaussian.
# @author: Ronald.
def pstfailure_primeq_F(x):
    return log(sinh(x)/x)


#Calculate upper bound on PST decryption/decapsulation failure probability, using two methods: Chernoff tail probability, and using an approximation of the (scaled) decryption error's distribution to a discrete Gaussian distribution.
def pstfailure_primeq(q,p,t,n,h,B,ring,key_size,is_hnf=False,is_binary=False,is_ternary=False,is_sparsehnf=False,is_spbinary=False,is_spternary=True):
    twoB = 2**B
    if(ring!=1):    #should not be calling this function if not in the NTT-friendly ring instantiation.
        print "Should not be calling pstfailure_primeq() for a non-NTT friendly ring instantiation. Abort."
        exit()
    x = var('a')
    mm = find_local_minimum( (4*h+1)*pstfailure_primeq_F(x/2) + (6*h+1)*pstfailure_primeq_F(x*p/2/q) + pstfailure_primeq_F(x*p/t) - p*x/2/twoB, 0.1, 5 )     #WARNING: find_local_minimum() requires SAGE.
    pbitchernov =  log(2*key_size*exp(mm[0]),2).n()
    v = ( (4*h) + 1 + ((6*h+1)*(p/q)**2) + ((p/t)**2) )/12
    pbitgauss = log(key_size*erfc( p/2/twoB/sqrt(2.0*v )), 2.0).n()        #taking the union bound over the desired keylength, not over n.
    return (pbitgauss,pbitchernov)    #return the PST failure probability (log_2) using the approximate discret Gaussian method (tighter/more accurate bound) and the Chernoff tail probability method (less accurate, more conservative).


def close_to_power_of_two(q,th):
    exp  = int(floor(log2(q)))
    power = 2**exp
    ok = 0
    if ((q < power+th) and (q > power - th)):
        ok = 1
    return ok


#Generate a dictionary/HashMap of the form KEY=prime q, VALUE=List of prime nplus1's satisfying q=1(mod nplus1).
def get_dict_qnplus1(nplus1_min,nplus1_max,q_min,q_max,th):
    qprimes_list=[]
    q=q_min
    while(q<=q_max):
        if(isprime(q)):
            qprimes_list.append(q)
        q=nextprime(q)
    #print "qprimes_list",qprimes_list
    dict_qm=defaultdict(list)
    for q in qprimes_list:
        nplus1=nplus1_min
        nplus1_list=[]
        while(nplus1<=nplus1_max):
            if(isprime(nplus1) and q%nplus1==1 and nplus1 not in nplus1_list and (close_to_power_of_two(q,th)==1)):
                nplus1_list.append(nplus1)
            nplus1=nextprime(nplus1)
        if nplus1_list:
            dict_qm[q]=nplus1_list
    #print "dict_qm",dict_qm
    if not dict_qm:
        print "Warning. Dictionary of (q,m[]) is empty. Check bounds for q and nplus1."
        return None
    else:
        return dict_qm


# Generate an exhaustive lookup table mapping PST parameters for the NTT Ring case, to corresponding F.P. when calculated using the approximation-to-discrete-Gaussian method.
# The lookup table is a Python dictionary, i.e., a HashMap with 
# keys in the form of a concatenated string q_p_t_n_h_B_keysize (simplifies hashing and lookups), we omit the B and ring parameter since for this method both are assumed to be 1,
# and values that are the corresponding F.P.
# Parameters:
# dict_q_nplus1 is a dictionary mapping prime q's to prime (n+1)'s that satisfy q=1(mod n+1),
# key_sizes is a list containing desired key-sizes,
# minroundbits is the minimum LWR noise that we wish to have (represented as number of bits dropped when going from q to p).
def gen_primeq_FP_table(dict_q_nplus1,theta_min,theta_max,theta_step,B,ring,key_sizes,minroundbits,is_hnf=False,is_binary=False,is_ternary=False,is_sparsehnf=False,is_spbinary=False,is_spternary=True):
    primeq_fp_dict = {}    #a Python dictionary aka HashMap.
    #Sanity check
    if(B!=1 or ring!=1):
        print "Failure probability lookup table should only be created for the NTT ring case, and B should be 1 here. Abort."
        exit()
    q_list = sorted(dict_q_nplus1.keys())
    logp_min = int(floor(log2(q_list[0]))) - minroundbits
    for key_size in key_sizes:
        for q in q_list:
	    logp_max = int(floor(log2(q))) - minroundbits
	    for logp in range(logp_min, logp_max+1):
		p = 2**logp
                for nplus1 in dict_q_nplus1[q]:
                    n = nplus1 - 1
                    theta = theta_min
                    while(theta <= theta_max):
                        hammwt = int(floor(theta*n))
                        if((hammwt%2) != 0):
                            hammwt -= 1    #we consider only even hamming weights
                        logt_min = B
                        logt_max = logp
                        for logt in range(logt_min, logt_max+1):
                            t = 2**logt
                            current_param_set = str(q)+"_"+str(p)+"_"+str(t)+"_"+str(n)+"_"+str(hammwt)+"_"+str(B)+"_"+str(ring)+"_"+str(key_size)
                            (chernoff_fail, approx_discGaussian_fail) = pstfailure_primeq(q,p,t,n,hammwt,B,ring,key_size)
                            #Replace weird SAGE data types, e.g., 'infinity' with Python data types
                            if(chernoff_fail < (-1*log_infinity)):
                                chernoff_fail = float(-1*log_infinity)
                            elif(chernoff_fail > log_infinity):
                                chernoff_fail = float(log_infinity)
                            if(approx_discGaussian_fail < (-1*log_infinity)):
                                approx_discGaussian_fail = float(-1*log_infinity)
                            elif(approx_discGaussian_fail > log_infinity):
                                approx_discGaussian_fail = float(log_infinity)
                            if(abs(chernoff_fail - approx_discGaussian_fail)>10):
                                primeq_fp_dict[current_param_set] = chernoff_fail
#                                print "Chose the Chernoff failure tail probability."
                            else:
                                primeq_fp_dict[current_param_set] = approx_discGaussian_fail
#                                print "Chose the approximation to a discrete Gaussian tail probability."
                        theta += theta_step
    if not primeq_fp_dict:
        print "Warning. Generated Parameter_set->F.P. lookup table is empty. Check bounds.."
        return None
    else:
        return primeq_fp_dict


# Lookup the failure probability for a given parameter set, calculated using an approximation to a discrete Gaussian distribution.
def lookup_primeq_fp(primeq_fp_lookuptable,q,p,t,n,hammwt,B,ring,key_size):
    if primeq_fp_lookuptable:
        key = str(q)+"_"+str(p)+"_"+str(t)+"_"+str(n)+"_"+str(hammwt)+"_"+str(B)+"_"+str(ring)+"_"+str(key_size)
        try:
            return primeq_fp_lookuptable[key]
        except KeyError:
            print "Lookup failed. No such parameter set exists in the F.P. lookup table."
            return log_infinity
    else:
        print "Failure probability lookup table is empty. Abort."
        return log_infinity


#Run it
n_min = 200
nplus1_min = n_min + 1
nplus1_max = 2000
q_min = 2**8
q_max = 2**15
th = 256

theta_min = 0.12
theta_max = 0.25
theta_step = 0.01
B = 1
ring = 1
key_sizes = [128,192,256,384]
minroundbits = 2

# Create the lookup table for PST_paramset->F.P. for the NTT ring case
pst_primeq_fp_lookuptable = gen_primeq_FP_table(get_dict_qnplus1(nplus1_min,nplus1_max,q_min,q_max,th),theta_min,theta_max,theta_step,B,ring,key_sizes,minroundbits)
print "\nCreated the PST Parameterset->F.P. lookup table. Writing to file now."
lt_filename = "/home/310220923/Software/nistpq/parameters/ntt_ringcase/pst_fp_lookuptable.txt"
lt_file = open(lt_filename, 'w')
lt_file.seek(0)    #start writing at beginning of file
lt_file.write(str(pst_primeq_fp_lookuptable))
lt_file.close()
print "Wrote PST Parameterset->F.P. lookup table to",lt_filename
